from collections import deque
from functools import reduce
from typing import Any, Callable, Dict, Iterable, List, Optional, TypeVar, Union

import immutables
from ordered_set import OrderedSet

T = TypeVar("T")
U = TypeVar("U")

NodeInput = Dict[str, List[Any]]  # Of format {node_address: []}
Row = Dict[str, Any]
FIDESOPS_DO_NOT_MASK_INDEX = "FIDESOPS_DO_NOT_MASK"


def make_immutable(obj: Any) -> Any:
    """
    Recursively converts a mutable object into an immutable version.
    Dictionaries are converted to immutable `Map`s from the `immutables` library,
    lists are converted to tuples, and other objects are returned unchanged.
    """
    if isinstance(obj, dict):
        return immutables.Map(
            {key: make_immutable(value) for key, value in obj.items()}
        )
    if isinstance(obj, list):
        return tuple(make_immutable(item) for item in obj)
    return obj


def make_mutable(obj: Any) -> Any:
    """
    Recursively converts an immutable object into a mutable version.
    `Map`s from the `immutables` library and dictionaries are converted to mutable dictionaries,
    tuples and `OrderedSet`s are converted to lists, and other objects are returned unchanged.
    """
    if isinstance(obj, (dict, immutables.Map)):
        return {key: make_mutable(value) for key, value in obj.items()}
    if isinstance(obj, (tuple, OrderedSet)):
        return [make_mutable(item) for item in obj]
    return obj


def merge_dicts(*dicts: Dict[T, U]) -> Dict[T, U]:
    """Merge any number of dictionaries.

    Right-hand values take precedence. That is,
    merge_dicts({"A": 1, "B": 2}, {"A": 2, "C": 4})
    =>  {'A': 2, 'B': 2, 'C': 4}
    """
    if dicts:
        return reduce(lambda x, y: {**x, **y}, dicts) or {}
    return {}


def append(d: Dict[T, List[U]], key: T, value: U) -> None:
    """Append to values stored under a dictionary key.

    append({},"A",1) sets dict to {"A":[1]}
    append({"A":[1],"A",2) sets dict to {"A":[1,2]}
    append({"A":[1],"A",[2, 3, 4]) sets dict to {"A":[1, 2, 3, 4]}
    """
    if value:
        if key in d:
            if isinstance(value, list):
                d[key].extend(value)
            else:
                d[key].append(value)
        else:
            d[key] = value if isinstance(value, list) else [value]


def append_unique(d: Dict[T, OrderedSet[U]], key: T, value: Union[U, List[U]]) -> None:
    """Append to values stored under a dictionary key.

    append_unique({}, "A", 1) sets dict to {"A": {1}}
    append_unique({"A": {1}}, "A", 2) sets dict to {"A": {1, 2}}
    append_unique({"A": {1}}, "A", [2, 3, 4]) sets dict to {"A": {1, 2, 3, 4}}
    append_unique({"A": {1, 2}}, "A", [2, 3, 4]) sets dict to {"A": {1, 2, 3, 4}}
    """
    if value:
        if key not in d:
            d[key] = OrderedSet()

        if isinstance(value, list):
            d[key].update(make_immutable(value))
        else:
            d[key].add(make_immutable(value))


def partition(_iterable: Iterable[T], extractor: Callable[[T], U]) -> Dict[U, List[T]]:
    """partition a collection by the output of an arbitrary extractor function"""
    out: Dict[U, List[T]] = {}
    for t in _iterable:
        append(out, extractor(t), t)
    return out


def filter_nonempty_values(d: Optional[Dict[Any, Any]]) -> Dict[Any, Any]:
    """Return the input map with empty values removed. On an input of None
    will return an empty Dict."""
    if d:
        return {e[0]: e[1] for e in d.items() if e[1]}
    return {}


def extract_key_for_address(
    full_request_id: str, number_of_leading_strings_to_exclude: int
) -> str:
    """
    Handles extracting the correct Dataset:Collection to map to extracted
    values.

    Due to differences in the number of leading strings based on access or
    erasure, a parameter is used to ensure the correct values are returned.

    Handles an edge case where double underscores exist in either the fides_key
    of the Dataset or the Collection name.
    """
    request_id_dataset, collection = full_request_id.split(":")
    dataset = request_id_dataset.split("__", number_of_leading_strings_to_exclude)[-1]
    return f"{dataset}:{collection}"


def unflatten_dict(flat_dict: Dict[str, Any], separator: str = ".") -> Dict[str, Any]:
    """
    Converts a dictionary of paths/values into a nested dictionary

    example:

    {"A.B": "1", "A.C": "2"}

    becomes

    {
        "A": {
            "B": "1",
            "C": "2"
        }
    }
    """
    output: Dict[Any, Any] = {}
    queue = deque(flat_dict.items())

    while queue:
        path, value = queue.popleft()
        keys = path.split(separator)
        target = output
        for i, current_key in enumerate(keys[:-1]):
            next_key = keys[i + 1]
            if next_key.isdigit():
                target = target.setdefault(current_key, [])
            else:
                if isinstance(target, dict):
                    target = target.setdefault(current_key, {})
                elif isinstance(target, list):
                    while len(target) <= int(current_key):
                        target.append({})
                    target = target[int(current_key)]
        try:
            if isinstance(target, list):
                target.append(value)
            else:
                # If the value is a dictionary, add its components to the queue for processing
                if isinstance(value, dict):
                    target = target.setdefault(keys[-1], {})
                    for inner_key, inner_value in value.items():
                        new_key = f"{path}{separator}{inner_key}"
                        queue.append((new_key, inner_value))
                else:
                    target[keys[-1]] = value
        except TypeError as exc:
            raise ValueError(
                f"Error unflattening dictionary, conflicting levels detected: {exc}"
            )
    return output


def flatten_dict(data: Any, prefix: str = "", separator: str = ".") -> Dict[str, Any]:
    """
    Recursively flatten a dictionary or list into a flat dictionary with dot-notation keys.
    Handles nested dictionaries and arrays with proper indices.

    example:

    {
        "A": {
            "B": "1",
            "C": "2"
        },
        "D": [
            {"E": "3"},
            {"E": "4"}
        ]
    }

    becomes

    {
        "A.B": "1",
        "A.C": "2",
        "D.0.E": "3",
        "D.1.E": "4"
    }

    Args:
        data: The data to flatten (dict, list, or scalar value)
        prefix: The current key prefix (used in recursion)
        separator: The separator to use between key segments (default: ".")

    Returns:
        A flattened dictionary with dot-notation keys
    """
    items = {}

    if isinstance(data, dict):
        for k, v in data.items():
            new_key = f"{prefix}{separator}{k}" if prefix else k
            if isinstance(v, (dict, list)):
                items.update(flatten_dict(v, new_key, separator))
            else:
                items[new_key] = v
    elif isinstance(data, list):
        for i, v in enumerate(data):
            new_key = f"{prefix}{separator}{i}"
            if isinstance(v, (dict, list)):
                items.update(flatten_dict(v, new_key, separator))
            else:
                items[new_key] = v
    else:
        raise ValueError(
            f"Input to flatten_dict must be a dict or list, got {type(data).__name__}"
        )

    return items


def replace_none_arrays(
    data: Any,
) -> Any:
    """
    Recursively replace any arrays containing only None values with empty arrays.

    Args:
        data: Any Python object (dict, list, etc.)

    Returns:
        The transformed data structure with None-only arrays replaced by empty arrays
    """
    if isinstance(data, dict):
        # Process each key-value pair in dictionaries
        return {k: replace_none_arrays(v) for k, v in data.items()}

    if isinstance(data, list):
        # Check if list contains only None values
        if all(item is None for item in data):
            return []

        # Otherwise process each item in the list
        return [replace_none_arrays(item) for item in data]

    # Return other data types unchanged
    return data



"""adds auto_subdomain_cookie_deletion on experience config, migrate to write initial val of false

Revision ID: 36ad82edb38e
Revises: 5dafbc1818ae
Create Date: 2024-10-07 19:30:03.559747

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "36ad82edb38e"
down_revision = "5dafbc1818ae"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "experienceconfigtemplate",
        sa.Column(
            "auto_subdomain_cookie_deletion",
            sa.Boolean(),
            server_default="t",
            nullable=False,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "auto_subdomain_cookie_deletion",
            sa.Boolean(),
            server_default="t",
            nullable=False,
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "auto_subdomain_cookie_deletion",
            sa.Boolean(),
            nullable=True,
        ),
    )

    # Updates all existing privacy experience configs to be False. Moving forward this will be set to True by default.
    op.execute(
        "UPDATE privacyexperienceconfig SET auto_subdomain_cookie_deletion = FALSE"
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacyexperienceconfig", "auto_subdomain_cookie_deletion")
    op.drop_column("experienceconfigtemplate", "auto_subdomain_cookie_deletion")
    op.drop_column("privacyexperienceconfighistory", "auto_subdomain_cookie_deletion")
    # ### end Alembic commands ###



"""Upgrade to fideslang 1.4

Revision ID: 63b482f5b49b
Revises: deb97d9393f3
Create Date: 2023-05-26 07:51:25.947974

"""

import json
from typing import Dict, List, Optional

from alembic import op
from loguru import logger
from sqlalchemy import text
from sqlalchemy.engine import Connection, ResultProxy
from sqlalchemy.sql.elements import TextClause

# revision identifiers, used by Alembic.
revision = "63b482f5b49b"
down_revision = "deb97d9393f3"
branch_labels = None
depends_on = None

# The `key` is the old value, the `value` is the new value
data_use_upgrades: Dict[str, str] = {
    "third_party_sharing.payment_processing": "essential.service.payment_processing",
    "third_party_sharing.fraud_detection": "essential.service.fraud_detection",
    "advertising": "marketing.advertising",
    "advertising.first_party": "marketing.advertising.first_party",
    "advertising.first_party.contextual": "marketing.advertising.first_party.contextual",
    "advertising.first_party.personalized": "marketing.advertising.first_party.targeted",
    "advertising.third_party": "marketing.advertising.third_party.targeted",
    "advertising.third_party.personalized": "marketing.advertising.third_party.targeted",
    "third_party_sharing.personalized_advertising": "marketing.advertising.third_party.targeted",
    "provide": "essential",
    "provide.service": "essential.service",
}
data_use_downgrades: Dict[str, str] = {
    value: key for key, value in data_use_upgrades.items()
}


def update_privacy_declaration_data_uses(
    bind: Connection, data_use_map: Dict[str, str]
):
    """Upgrade or downgrade data uses from fideslang 1.4 for privacy declarations"""
    existing_ctl_policies: ResultProxy = bind.execute(
        text("SELECT id, data_use FROM privacydeclaration;")
    )
    for row in existing_ctl_policies:
        data_use: str = row["data_use"]
        updated_data_use: Optional[str] = data_use_map.get(data_use, None)

        if updated_data_use:
            update_data_use_query: TextClause = text(
                "UPDATE privacydeclaration SET data_use = :updated_use WHERE id= :declaration_id"
            )
            bind.execute(
                update_data_use_query,
                {"declaration_id": row["id"], "updated_use": updated_data_use},
            )


def update_ctl_policy_data_uses(bind: Connection, data_use_map: Dict[str, str]):
    """Upgrade or downgrade data uses from fideslang 1.4 for ctl policies"""
    existing_ctl_policies: ResultProxy = bind.execute(
        text("SELECT id, rules FROM ctl_policies;")
    )
    for row in existing_ctl_policies:
        needs_update: bool = False
        rules: List[Dict] = row["rules"]
        for i, rule in enumerate(rules or []):
            data_uses: Dict = rule.get("data_uses", {})
            for j, val in enumerate(data_uses.get("values", [])):
                new_data_use: Optional[str] = data_use_map.get(val, None)
                if new_data_use:
                    rules[i]["data_uses"]["values"][j] = new_data_use
                    needs_update = True

        if needs_update:
            update_data_use_query: TextClause = text(
                "UPDATE ctl_policies SET rules = :updated_rules WHERE id= :policy_id"
            )
            bind.execute(
                update_data_use_query,
                {"policy_id": row["id"], "updated_rules": json.dumps(rules)},
            )


def upgrade() -> None:
    bind: Connection = op.get_bind()

    logger.info("Upgrading data use on privacy declaration")
    update_privacy_declaration_data_uses(bind, data_use_upgrades)

    logger.info("Upgrading ctl policy rule data uses")
    update_ctl_policy_data_uses(bind, data_use_upgrades)


def downgrade():
    bind: Connection = op.get_bind()

    logger.info("Downgrading data use on privacy declaration")
    update_privacy_declaration_data_uses(bind, data_use_downgrades)

    logger.info("Downgrading ctl policy rule data uses")
    update_ctl_policy_data_uses(bind, data_use_downgrades)



from enum import Enum
from typing import Any, Iterable, List, Optional

from pydantic import field_validator

from fides.api.custom_types import CssStr
from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.privacy_center_config import PrivacyCenterConfig


class MinimalPrivacyExperienceConfig(FidesSchema):
    """
    Minimal representation of a privacy experience config, contains enough information
    to select experience configs by name in the UI and an ID to link the selections in the database.

    NOTE: Add to this schema with care. Any fields added to
    this response schema will be exposed in public-facing
    (i.e. unauthenticated) API responses. If a field has
    sensitive information, it should NOT be added to this schema!
    """

    id: str
    name: str


class MinimalMessagingTemplate(FidesSchema):
    """
    Minimal representation of a messaging template.

    NOTE: Add to this schema with care. Any fields added to
    this response schema will be exposed in public-facing
    (i.e. unauthenticated) API responses. If a field has
    sensitive information, it should NOT be added to this schema!
    """

    id: str
    type: str


class PropertyType(Enum):
    website = "Website"
    other = "Other"


class MinimalProperty(FidesSchema):
    id: str
    name: str


class PropertyCreate(FidesSchema):
    name: str
    type: PropertyType
    id: Optional[str] = None
    experiences: List[MinimalPrivacyExperienceConfig]
    privacy_center_config: Optional[PrivacyCenterConfig] = None
    stylesheet: Optional[CssStr] = None
    paths: List[str]


class PublicPropertyResponse(FidesSchema):
    """
    Schema that represents a `Property` as returned in the
    public-facing Property APIs.

    NOTE: Add to this schema with care. Any fields added to
    this response schema will be exposed in public-facing
    (i.e. unauthenticated) API responses. If a field has
    sensitive information, it should NOT be added to this schema!

    Any `Property` fields that are sensitive should be added to the
    appropriate non-public schemas that extend this schema.
    """

    name: str
    type: PropertyType
    id: Optional[str] = None
    experiences: List[MinimalPrivacyExperienceConfig]
    messaging_templates: Optional[List[MinimalMessagingTemplate]] = None
    privacy_center_config: Optional[PrivacyCenterConfig] = None
    stylesheet: Optional[CssStr] = None
    paths: List[str]

    @field_validator("paths", mode="before")
    @classmethod
    def convert_to_list(cls, value: Any) -> Any:  # type: ignore[misc]
        """
        Convert the 'paths' value to a list if it is an iterable of strings.
        This validator is necessary because SQLAlchemy returns the 'paths' value
        as an iterable (association proxy) instead of a list. The validator checks
        if the 'paths' value is an iterable (excluding strings) and if all its
        elements are strings. If these conditions are met, it converts the iterable
        to a list. Otherwise, it returns the original value unchanged.
        """
        if (
            isinstance(value, Iterable)
            and not isinstance(value, str)
            and all(isinstance(item, str) for item in value)
        ):
            return list(value)
        return value


class Property(PublicPropertyResponse):
    """
    A schema representing the complete `Property` model.

    This schema extends the base `PublicPropertyResponse` schema,
    which only includes fields that are appropriate to be exposed
    in public endpoints.

    Any `Property` fields that are sensitive but need to be included in private
    API responses should be added to this schema.
    """



"""Add DBCache table

Revision ID: e5ec30dfcd87
Revises: c90d46f6d3f2
Create Date: 2024-12-17 16:48:04.006190

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "e5ec30dfcd87"
down_revision = "c90d46f6d3f2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "dbcache",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("namespace", sa.String(), nullable=False),
        sa.Column("cache_key", sa.String(), nullable=False),
        sa.Column("cache_value", postgresql.BYTEA(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_dbcache_id"), "dbcache", ["id"], unique=False)
    op.create_index(
        op.f("ix_dbcache_namespace"), "dbcache", ["namespace"], unique=False
    )
    # Index over (namespace, cache_key), with unique set to True to ensure
    # unique cache keys within a namespace
    op.create_index(
        op.f("ix_dbcache_namespace_cache_key"),
        "dbcache",
        ["namespace", "cache_key"],
        unique=True,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_dbcache_namespace"), table_name="dbcache")
    op.drop_index(op.f("ix_dbcache_id"), table_name="dbcache")
    op.drop_index(op.f("ix_dbcache_namespace_cache_key"), table_name="dbcache")
    op.drop_table("dbcache")
    # ### end Alembic commands ###



from json import loads

from sqlalchemy import create_engine, inspect
from sqlalchemy.engine import Engine

from fides.connectors.models import BigQueryConfig, ConnectorFailureException


def get_bigquery_engine(bigquery_config: BigQueryConfig) -> Engine:
    """
    Creates the SQLAlchemy engine for connecting to a BigQuery dataset
    """
    project_id = bigquery_config.keyfile_creds.project_id
    dataset = bigquery_config.dataset
    engine = create_engine(
        f"bigquery://{project_id}/{dataset}",
        credentials_info=bigquery_config.keyfile_creds.model_dump(mode="json"),
    )
    validate_bigquery_engine(engine)
    return engine


def validate_bigquery_engine(engine: Engine) -> None:
    """
    Tests the BigQuery engine by calling the schema inspection

    Using the .connect() method does not produce an exception
    """
    from google.api_core.exceptions import ClientError

    inspector = inspect(engine)
    try:
        inspector.get_schema_names()
    except ClientError as error:
        raise ConnectorFailureException(loads(error.response.text)["error"]["message"])



from typing import ClassVar, List

from pydantic import Field

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets_base_aws import (
    BaseAWSSchema,
)


class DynamoDBSchema(BaseAWSSchema):
    """Schema to validate the secrets needed to connect to an Amazon DynamoDB cluster"""

    region_name: str = Field(
        title="Region",
        description="The AWS region where your DynamoDB table is located (ex. us-west-2).",
    )

    _required_components: ClassVar[List[str]] = BaseAWSSchema._required_components + [
        "region_name"
    ]


class DynamoDBDocsSchema(DynamoDBSchema, NoValidationSchema):
    """DynamoDB Secrets Schema for API Docs"""



"""disconnect_privacy_declaration_cookies_from_system

Revision ID: b3b407ce2d5f
Revises: ddec29f24945
Create Date: 2023-11-03 18:28:36.416814

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy import text

revision = "b3b407ce2d5f"
down_revision = "ddec29f24945"
branch_labels = None
depends_on = None


def upgrade():
    # Remove FK and add new one that cascade deletes. When Privacy Declaration is deleted, let's delete the Cookie.
    # Previously, we left it pointed to the System
    op.drop_constraint(
        "cookies_privacy_declaration_id_fkey", "cookies", type_="foreignkey"
    )
    op.create_foreign_key(
        "cookies_privacy_declaration_id_fkey",
        "cookies",
        "privacydeclaration",
        ["privacy_declaration_id"],
        ["id"],
        ondelete="CASCADE",
    )

    bind = op.get_bind()
    # Update existing cookies that are pointed to both a privacy declaration and a system id to just be
    # pointed to the privacy declaration id
    bind.execute(
        text(
            "UPDATE cookies SET system_id = null WHERE privacy_declaration_id IS NOT NULL and system_id IS NOT NULL;"
        )
    )


def downgrade():
    op.drop_constraint(
        "cookies_privacy_declaration_id_fkey", "cookies", type_="foreignkey"
    )
    op.create_foreign_key(
        "cookies_privacy_declaration_id_fkey",
        "cookies",
        "privacydeclaration",
        ["privacy_declaration_id"],
        ["id"],
        ondelete="SET NULL",
    )



# pylint: disable=R0401, C0302

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Optional

from loguru import logger
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, String, UniqueConstraint
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.mutable import MutableList
from sqlalchemy.orm import Session, relationship

from fides.api.db.base_class import Base  # type: ignore[attr-defined]
from fides.api.db.util import EnumColumn
from fides.api.models.privacy_request.privacy_request import CustomPrivacyRequestField
from fides.api.models.privacy_request.provided_identity import ProvidedIdentity
from fides.api.schemas.privacy_request import PrivacyRequestSource
from fides.api.schemas.redis_cache import (
    CustomPrivacyRequestField as CustomPrivacyRequestFieldSchema,
)
from fides.api.schemas.redis_cache import IdentityBase
from fides.api.util.cache import FidesopsRedis, get_cache
from fides.api.util.identity_verification import IdentityVerificationMixin
from fides.config import CONFIG

if TYPE_CHECKING:
    from fides.api.models.privacy_request.privacy_request import (
        PrivacyRequest,  # type: ignore[attr-defined]
    )


class Consent(Base):
    """The DB ORM model for Consent."""

    provided_identity_id = Column(
        String,
        ForeignKey(ProvidedIdentity.id),
        nullable=False,
    )
    data_use = Column(String, nullable=False)
    data_use_description = Column(String)
    opt_in = Column(Boolean, nullable=False)
    has_gpc_flag = Column(
        Boolean,
        server_default="f",
        default=False,
        nullable=False,
    )
    conflicts_with_gpc = Column(
        Boolean,
        server_default="f",
        default=False,
        nullable=False,
    )

    provided_identity = relationship(ProvidedIdentity, back_populates="consent")

    UniqueConstraint(provided_identity_id, data_use, name="uix_identity_data_use")

    identity: Optional[IdentityBase] = None


class ConsentRequest(IdentityVerificationMixin, Base):
    """Tracks consent requests."""

    property_id = Column(
        String,
        nullable=True,
    )
    provided_identity_id = Column(
        String, ForeignKey(ProvidedIdentity.id), nullable=False
    )
    provided_identity = relationship(
        ProvidedIdentity,
        back_populates="consent_request",
    )

    custom_fields = relationship(
        "CustomPrivacyRequestField", back_populates="consent_request"
    )

    preferences = Column(
        MutableList.as_mutable(JSONB),
        nullable=True,
    )

    identity_verified_at = Column(
        DateTime(timezone=True),
        nullable=True,
    )

    source = Column(EnumColumn(PrivacyRequestSource), nullable=True)

    privacy_request_id = Column(String, ForeignKey("privacyrequest.id"), nullable=True)
    privacy_request = relationship("PrivacyRequest")

    def get_cached_identity_data(self) -> Dict[str, Any]:
        """Retrieves any identity data pertaining to this request from the cache."""
        prefix = f"id-{self.id}-identity-*"
        cache: FidesopsRedis = get_cache()
        keys = cache.keys(prefix)
        return {key.split("-")[-1]: cache.get(key) for key in keys}

    def verify_identity(
        self,
        db: Session,
        provided_code: Optional[str] = None,
    ) -> None:
        """
        A method to call the internal identity verification method provided by the
        `IdentityVerificationMixin`.
        """
        self._verify_identity(provided_code=provided_code)
        self.identity_verified_at = datetime.utcnow()
        self.save(db)

    def persist_custom_privacy_request_fields(
        self,
        db: Session,
        custom_privacy_request_fields: Optional[
            Dict[str, CustomPrivacyRequestFieldSchema]
        ],
    ) -> None:
        if not custom_privacy_request_fields:
            return

        if CONFIG.execution.allow_custom_privacy_request_field_collection:
            for key, item in custom_privacy_request_fields.items():
                if item.value:
                    hashed_value = CustomPrivacyRequestField.hash_value(item.value)
                    CustomPrivacyRequestField.create(
                        db=db,
                        data={
                            "consent_request_id": self.id,
                            "field_name": key,
                            "field_label": item.label,
                            "encrypted_value": {"value": item.value},
                            "hashed_value": hashed_value,
                        },
                    )
        else:
            logger.info(
                "Custom fields provided in consent request {}, but config setting 'CONFIG.execution.allow_custom_privacy_request_field_collection' prevents their storage.",
                self.id,
            )

    def get_persisted_custom_privacy_request_fields(self) -> Dict[str, Any]:
        return {
            field.field_name: {
                "label": field.field_label,
                "value": field.encrypted_value["value"],
            }
            for field in self.custom_fields  # type: ignore[attr-defined]
        }



"""add openid_provider

Revision ID: ffee79245c9a
Revises: d69cf8f82a58
Create Date: 2024-08-09 19:27:07.222226

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "ffee79245c9a"
down_revision = "d69cf8f82a58"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "openid_provider",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("identifier", sa.String(), nullable=True),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column(
            "provider",
            sa.Enum("google", "okta", "custom", name="providerenum"),
            nullable=True,
        ),
        sa.Column(
            "client_id",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=False,
        ),
        sa.Column(
            "client_secret",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=False,
        ),
        sa.Column("domain", sa.String(), nullable=True),
        sa.Column("authorization_url", sa.String(), nullable=True),
        sa.Column("token_url", sa.String(), nullable=True),
        sa.Column("user_info_url", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_openid_provider_id"), "openid_provider", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_openid_provider_identifier"),
        "openid_provider",
        ["identifier"],
        unique=True,
    )


def downgrade():
    op.drop_index(op.f("ix_openid_provider_identifier"), table_name="openid_provider")
    op.drop_index(op.f("ix_openid_provider_id"), table_name="openid_provider")
    op.drop_table("openid_provider")
    op.execute("DROP TYPE providerenum")



import json
from datetime import date, datetime
from enum import Enum
from typing import Any, Dict
from urllib.parse import quote, unquote_to_bytes

from bson.objectid import ObjectId

ENCODED_BYTES_PREFIX = "quote_encoded_"
ENCODED_DATE_PREFIX = "date_encoded_"
ENCODED_MONGO_OBJECT_ID_PREFIX = "encoded_object_id_"


class CustomJSONEncoder(json.JSONEncoder):
    """
    CustomJSONEncoder that extends the builtin JSONEncoder,
    used to handle serialization of data types not supported by
    json.dumps().

    This encoder can be used for JSON serialization for storage in
    the redis cache, as well as our DB JSONB columns/fields.
    """

    def default(self, o: Any) -> Any:  # pylint: disable=too-many-return-statements
        if isinstance(o, Enum):
            return o.value
        if isinstance(o, bytes):
            return f"{ENCODED_BYTES_PREFIX}{quote(o)}"
        if isinstance(o, (datetime, date)):
            return f"{ENCODED_DATE_PREFIX}{o.isoformat()}"
        if isinstance(o, ObjectId):
            return f"{ENCODED_MONGO_OBJECT_ID_PREFIX}{str(o)}"
        if isinstance(o, object):
            if hasattr(o, "__dict__"):
                return o.__dict__
            if not isinstance(o, int) and not isinstance(o, float):
                return str(o)

        # It doesn't seem possible to make it here, but I'm leaving in as a fail safe
        # just in case.
        return super().default(o)  # pragma: no cover


def _custom_decoder(json_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Custom JSON decoder that deserializes data that's been serialized,
    by the CustomJSONEncoder, which is used used to handle serialization
    of data types not supported by json.dumps().

    This encoder can be used for JSON deserialization for storage in
    the redis cache, as well as our DB JSONB columns/fields.
    """
    for k, v in json_dict.items():
        if isinstance(v, str):
            # The mongodb objectids couldn't be directly json encoded so they are converted
            # to strings and prefixed with encoded_object_id in order to find during decodeint.
            if v.startswith(ENCODED_MONGO_OBJECT_ID_PREFIX):
                json_dict[k] = ObjectId(v[18:])
            if v.startswith(ENCODED_DATE_PREFIX):
                json_dict[k] = datetime.fromisoformat(v[13:])
            # The bytes from secrets couldn't be directly json encoded so it is url
            # encode and prefixed with quite_encoded in order to find during decodeint.
            elif v.startswith(ENCODED_BYTES_PREFIX):
                json_dict[k] = unquote_to_bytes(v)[14:]

    return json_dict



from typing import ClassVar, List, Optional

from pydantic import Field, model_validator

from fides.api.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)
from fides.api.schemas.storage.storage import AWSAuthMethod


class BaseAWSSchema(ConnectionConfigSecretsSchema):
    """Base schema for AWS connection secrets"""

    auth_method: AWSAuthMethod = Field(
        title="Authentication Method",
        description="Determines which type of authentication method to use for connecting to Amazon Web Services. Currently accepted values are: `secret_keys` or `automatic`.",
    )

    aws_access_key_id: Optional[str] = Field(
        default=None,
        title="Access Key ID",
        description="Part of the credentials that provide access to your AWS account.",
    )
    aws_secret_access_key: Optional[str] = Field(
        default=None,
        title="Secret Access Key",
        description="Part of the credentials that provide access to your AWS account.",
        json_schema_extra={"sensitive": True},
    )

    aws_assume_role_arn: Optional[str] = Field(
        default=None,
        title="Assume Role ARN",
        description="If provided, the ARN of the role that should be assumed to connect to AWS.",
    )

    _required_components: ClassVar[List[str]] = ["auth_method"]

    @model_validator(mode="after")
    def keys_provided_if_needed(self) -> "BaseAWSSchema":
        """
        Validates that both access and secret access keys are provided if using a `secret_keys` auth method.
        """
        if self.auth_method == AWSAuthMethod.SECRET_KEYS.value and not (
            self.aws_access_key_id and self.aws_secret_access_key
        ):
            raise ValueError(
                f"An Access Key ID and a Secret Access Key must be provided if using the `{AWSAuthMethod.SECRET_KEYS.value}` Authentication Method"
            )

        return self



from typing import Any, List

from fastapi import HTTPException, Security
from loguru import logger
from starlette.status import HTTP_400_BAD_REQUEST, HTTP_404_NOT_FOUND

from fides.api.common_exceptions import NoSuchStrategyException, ValidationError
from fides.api.oauth.utils import verify_oauth_client_prod
from fides.api.schemas.masking.masking_api import MaskingAPIRequest, MaskingAPIResponse
from fides.api.schemas.masking.masking_strategy_description import (
    MaskingStrategyDescription,
)
from fides.api.schemas.policy import PolicyMaskingSpec
from fides.api.service.masking.strategy.masking_strategy import MaskingStrategy
from fides.api.util.api_router import APIRouter
from fides.common.api.scope_registry import MASKING_EXEC, MASKING_READ
from fides.common.api.v1.urn_registry import MASKING, MASKING_STRATEGY, V1_URL_PREFIX

router = APIRouter(tags=["Masking"], prefix=V1_URL_PREFIX)


@router.put(
    MASKING,
    response_model=MaskingAPIResponse,
    dependencies=[
        Security(
            verify_oauth_client_prod,
            scopes=[MASKING_EXEC],
        )
    ],
)
def mask_value(request: MaskingAPIRequest) -> MaskingAPIResponse:
    """Masks the value(s) provided using the provided masking strategy"""
    try:
        values: List[Any] = request.values
        masked_values: List[Any] = request.values.copy()
        masking_strategies: List[PolicyMaskingSpec] = request.masking_strategies

        num_strat: int = len(masking_strategies)

        if num_strat > 1:
            logger.info(
                "{} masking strategies requested; running in order.",
                num_strat,
            )

        for strategy in masking_strategies:
            masking_strategy = MaskingStrategy.get_strategy(
                strategy.strategy, strategy.configuration
            )
            logger.info(
                "Starting masking of {} value(s) with strategy {}",
                len(values),
                strategy.strategy,
            )
            masked_values = masking_strategy.mask(  # type: ignore
                masked_values, None
            )  # passing in masked values from previous strategy

        return MaskingAPIResponse(plain=values, masked_values=masked_values)
    except NoSuchStrategyException as e:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail=str(e))
    except ValidationError as e:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=str(e))
    except ValueError as e:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=str(e))


@router.get(
    MASKING_STRATEGY,
    response_model=List[MaskingStrategyDescription],
    dependencies=[
        Security(
            verify_oauth_client_prod,
            scopes=[MASKING_READ],
        )
    ],
)
def list_masking_strategies() -> List[MaskingStrategyDescription]:
    """Lists available masking strategies with instructions on how to use them"""
    logger.info("Getting available masking strategies")
    return [strategy.get_description() for strategy in MaskingStrategy.get_strategies()]



import pathlib

from jinja2 import Environment, FileSystemLoader, Template, select_autoescape
from loguru import logger

from fides.api.common_exceptions import EmailTemplateUnhandledActionType
from fides.api.email_templates.template_names import (
    CONSENT_REQUEST_EMAIL_FULFILLMENT,
    CONSENT_REQUEST_VERIFICATION_TEMPLATE,
    EMAIL_ERASURE_REQUEST_FULFILLMENT,
    PRIVACY_REQUEST_COMPLETE_ACCESS_TEMPLATE,
    PRIVACY_REQUEST_COMPLETE_DELETION_TEMPLATE,
    PRIVACY_REQUEST_ERROR_NOTIFICATION_TEMPLATE,
    PRIVACY_REQUEST_RECEIPT_TEMPLATE,
    PRIVACY_REQUEST_REVIEW_APPROVE_TEMPLATE,
    PRIVACY_REQUEST_REVIEW_DENY_TEMPLATE,
    SUBJECT_IDENTITY_VERIFICATION_TEMPLATE,
    TEST_MESSAGE_TEMPLATE,
    USER_INVITE,
)
from fides.api.schemas.messaging.messaging import MessagingActionType

pathlib.Path(__file__).parent.resolve()

abs_path_to_current_file_dir = pathlib.Path(__file__).parent.resolve()
template_env = Environment(
    loader=FileSystemLoader(f"{abs_path_to_current_file_dir}/templates"),
    autoescape=select_autoescape(),
)


def get_email_template(  # pylint: disable=too-many-return-statements
    action_type: MessagingActionType,
) -> Template:
    if action_type == MessagingActionType.CONSENT_REQUEST:
        return template_env.get_template(CONSENT_REQUEST_VERIFICATION_TEMPLATE)
    if action_type == MessagingActionType.SUBJECT_IDENTITY_VERIFICATION:
        return template_env.get_template(SUBJECT_IDENTITY_VERIFICATION_TEMPLATE)
    if action_type == MessagingActionType.MESSAGE_ERASURE_REQUEST_FULFILLMENT:
        return template_env.get_template(EMAIL_ERASURE_REQUEST_FULFILLMENT)
    if action_type == MessagingActionType.CONSENT_REQUEST_EMAIL_FULFILLMENT:
        return template_env.get_template(CONSENT_REQUEST_EMAIL_FULFILLMENT)
    if action_type == MessagingActionType.PRIVACY_REQUEST_RECEIPT:
        return template_env.get_template(PRIVACY_REQUEST_RECEIPT_TEMPLATE)
    if action_type == MessagingActionType.PRIVACY_REQUEST_COMPLETE_ACCESS:
        return template_env.get_template(PRIVACY_REQUEST_COMPLETE_ACCESS_TEMPLATE)
    if action_type == MessagingActionType.PRIVACY_REQUEST_COMPLETE_DELETION:
        return template_env.get_template(PRIVACY_REQUEST_COMPLETE_DELETION_TEMPLATE)
    if action_type == MessagingActionType.PRIVACY_REQUEST_ERROR_NOTIFICATION:
        return template_env.get_template(PRIVACY_REQUEST_ERROR_NOTIFICATION_TEMPLATE)
    if action_type == MessagingActionType.PRIVACY_REQUEST_REVIEW_DENY:
        return template_env.get_template(PRIVACY_REQUEST_REVIEW_DENY_TEMPLATE)
    if action_type == MessagingActionType.PRIVACY_REQUEST_REVIEW_APPROVE:
        return template_env.get_template(PRIVACY_REQUEST_REVIEW_APPROVE_TEMPLATE)
    if action_type == MessagingActionType.TEST_MESSAGE:
        return template_env.get_template(TEST_MESSAGE_TEMPLATE)
    if action_type == MessagingActionType.USER_INVITE:
        return template_env.get_template(USER_INVITE)

    logger.error("No corresponding template linked to the {}", action_type)
    raise EmailTemplateUnhandledActionType(
        f"No corresponding template linked to the {action_type}"
    )



"""Add last_run_timestamp to ConnectionConfig

Revision ID: 5a4859f74832
Revises: c9a22b284afa
Create Date: 2024-10-22 14:58:09.174708

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5a4859f74832"
down_revision = "c9a22b284afa"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "connectionconfig",
        sa.Column("last_run_timestamp", sa.DateTime(timezone=True), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("connectionconfig", "last_run_timestamp")
    # ### end Alembic commands ###



"""
Contains all of the endpoints required to manage generating resources.
"""

from enum import Enum
from typing import Dict, List, Optional, Union

from fastapi import Depends, HTTPException, Security, status
from fideslang.models import Dataset, Organization, System
from loguru import logger as log
from pydantic import BaseModel, model_validator
from sqlalchemy.ext.asyncio import AsyncSession

from fides.api.api.v1.endpoints import API_PREFIX
from fides.api.db.crud import get_resource
from fides.api.db.ctl_session import get_async_db
from fides.api.models.sql_models import sql_model_map  # type: ignore[attr-defined]
from fides.api.oauth.utils import verify_oauth_client_prod
from fides.api.util.api_router import APIRouter
from fides.common.api import scope_registry
from fides.connectors.models import (
    AWSConfig,
    BigQueryConfig,
    ConnectorAuthFailureException,
    ConnectorFailureException,
    DatabaseConfig,
    OktaConfig,
)
from fides.core.dataset import (
    generate_bigquery_datasets,
    generate_db_datasets,
    generate_dynamo_db_datasets,
)
from fides.core.system import generate_aws_systems, generate_okta_systems
from fides.core.utils import validate_db_engine

GENERATE_ROUTER = APIRouter(tags=["Generate"], prefix=f"{API_PREFIX}/generate")


class ValidTargets(str, Enum):
    """
    Validation of targets attempted to generate resources from
    """

    AWS = "aws"
    DB = "db"
    OKTA = "okta"
    BIGQUERY = "bigquery"
    DYNAMODB = "dynamodb"
    SCYLLADB = "scylla"


class GenerateTypes(str, Enum):
    """
    Generate Type Enum to capture the discrete possible values
    for a valid type of resource to generate.
    """

    SYSTEM = "systems"
    DATASET = "datasets"


class Generate(BaseModel):
    """
    Defines attributes for generating resources included in a request.
    """

    config: Union[AWSConfig, OktaConfig, DatabaseConfig, BigQueryConfig]
    target: ValidTargets
    type: GenerateTypes

    @model_validator(mode="before")
    @classmethod
    def target_matches_type(cls, values: Dict) -> Dict:
        """
        Ensures that both of the target and type attributes are a valid
        pair (returning an error on an ('aws', 'dataset') as an example).
        """
        target_type = (values.get("target"), values.get("type"))
        valid_target_types = [
            ("aws", "systems"),
            ("bigquery", "datasets"),
            ("db", "datasets"),
            ("dynamodb", "datasets"),
            ("okta", "systems"),
        ]
        if target_type not in valid_target_types:
            raise ValueError("Target and Type are not a valid match")
        return values


class GenerateRequestPayload(BaseModel):
    """
    The model for the request body housing generate information.
    """

    organization_key: str
    generate: Generate


class GenerateResponse(BaseModel):
    """
    The model to house the response for generated infrastructure.
    """

    generate_results: Optional[List[Union[Dataset, System]]] = None


@GENERATE_ROUTER.post(
    "/",
    status_code=status.HTTP_200_OK,
    dependencies=[
        Security(verify_oauth_client_prod, scopes=[scope_registry.GENERATE_EXEC])
    ],
    response_model=GenerateResponse,
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "content": {
                "application/json": {
                    "example": {"detail": "Unable to generate resources."}
                }
            }
        },
    },
)
async def generate(
    generate_request_payload: GenerateRequestPayload,
    db: AsyncSession = Depends(get_async_db),
) -> GenerateResponse:
    """
    A multi-purpose endpoint to generate Fides resources based on existing
    infrastructure

    Currently generates Fides resources for the following:
    * AWS: Systems
    * Okta: Systems
    * DB: Datasets
    * BigQuery: Datasets
    * DynamoDB: Datasets

    In the future, this will include options for other Systems & Datasets,
    examples include:
    * Snowflake: Datasets

    All production deployments should implement HTTPS for security purposes
    """
    organization: Organization = await get_resource(
        sql_model_map["organization"], generate_request_payload.organization_key, db
    )
    generate_config = generate_request_payload.generate.config
    generate_target = generate_request_payload.generate.target.lower()
    generate_results = None
    try:
        if generate_target == "aws" and isinstance(generate_config, AWSConfig):
            generate_results = generate_aws(
                aws_config=generate_config,
                organization=organization,
            )

        elif generate_target == "db" and isinstance(generate_config, DatabaseConfig):
            generate_results = generate_db(
                db_config=generate_config,
            )

        elif generate_target == "okta" and isinstance(generate_config, OktaConfig):
            generate_results = await generate_okta(
                okta_config=generate_config,
                organization=organization,
            )

        elif generate_target == "bigquery" and isinstance(
            generate_config, BigQueryConfig
        ):
            generate_results = generate_bigquery(
                bigquery_config=generate_config,
            )

        elif generate_target == "dynamodb" and isinstance(generate_config, AWSConfig):
            generate_results = generate_dynamodb(
                aws_config=generate_config,
                organization=organization,
                single_dataset=True,
            )

    except ConnectorAuthFailureException as error:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=str(error),
        )

    return GenerateResponse(generate_results=generate_results)


def generate_aws(
    aws_config: AWSConfig, organization: Organization
) -> List[Dict[str, str]]:
    """
    Returns a list of Systems found in AWS.
    """
    from fides.connectors.aws import validate_credentials

    log.info("Validating AWS credentials")
    try:
        validate_credentials(aws_config)
    except ConnectorAuthFailureException as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=str(error),
        )

    log.info("Generating systems from AWS")
    aws_systems = generate_aws_systems(organization=organization, aws_config=aws_config)

    return [i.model_dump(exclude_none=True) for i in aws_systems]


def generate_dynamodb(
    aws_config: AWSConfig, organization: Organization, single_dataset: bool
) -> List[Dict[str, str]]:
    """
    Returns a list of DynamoDB datasets found in AWS.
    """
    from fides.connectors.aws import validate_credentials

    log.info("Validating AWS credentials")
    try:
        validate_credentials(aws_config)
    except ConnectorAuthFailureException as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=str(error),
        )

    log.info("Generating datasets from AWS DynamoDB")
    aws_resources = generate_dynamo_db_datasets(
        aws_config=aws_config, single_dataset=single_dataset
    )

    return [i.model_dump(exclude_none=True) for i in aws_resources]


async def generate_okta(
    okta_config: OktaConfig, organization: Organization
) -> List[Dict[str, str]]:
    """
    Returns a list of Systems found in Okta.
    """
    from fides.connectors.okta import validate_credentials

    log.info("Validating Okta credentials")
    try:
        validate_credentials(okta_config)
    except ConnectorAuthFailureException as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=str(error),
        )
    log.info("Generating systems from Okta")
    okta_systems = await generate_okta_systems(
        organization=organization, okta_config=okta_config
    )
    return [i.model_dump(exclude_none=True) for i in okta_systems]


def generate_db(db_config: DatabaseConfig) -> List[Dict[str, str]]:
    """
    Returns a list of datasets found in a database.
    """
    log.info("Validating database credentials")
    try:
        validate_db_engine(db_config.connection_string)
    except ConnectorAuthFailureException as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=str(error),
        )
    log.info("Generating datasets from database")
    db_datasets = generate_db_datasets(connection_string=db_config.connection_string)

    return [i.model_dump(exclude_none=True) for i in db_datasets]


def generate_bigquery(bigquery_config: BigQueryConfig) -> List[Dict[str, str]]:
    """
    Returns a list of datasets found in a BigQuery dataset
    """
    log.info("Generating datasets from BigQuery")
    try:
        bigquery_datasets = generate_bigquery_datasets(bigquery_config)
    except ConnectorFailureException as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=str(error),
        )
    return [i.model_dump(exclude_none=True) for i in bigquery_datasets]



from typing import Any, Dict, List, Optional
from urllib.parse import quote_plus

from loguru import logger
from pymongo import MongoClient
from pymongo.errors import OperationFailure, ServerSelectionTimeoutError

from fides.api.common_exceptions import ConnectionException
from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest, RequestTask
from fides.api.schemas.connection_configuration.connection_secrets_mongodb import (
    MongoDBSchema,
)
from fides.api.service.connectors.base_connector import BaseConnector
from fides.api.service.connectors.query_configs.mongodb_query_config import (
    MongoQueryConfig,
)
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.util.collection_util import Row
from fides.api.util.logger import Pii


class MongoDBConnector(BaseConnector[MongoClient]):
    """MongoDB Connector"""

    def build_uri(self) -> str:
        """
        Builds URI of format mongodb://[username:password@]host1[:port1][,...hostN[:portN]][/[defaultauthdb][?options]]
        """

        config = MongoDBSchema(**self.configuration.secrets or {})

        user_pass: str = ""
        default_auth_db: str = ""
        if config.username and config.password:
            user_pass = f"{quote_plus(config.username)}:{quote_plus(config.password)}@"
            if config.defaultauthdb:
                default_auth_db = f"/{config.defaultauthdb}"

        port: str = f":{config.port}" if config.port else ""
        url = f"mongodb://{user_pass}{config.host}{port}{default_auth_db}"
        return url

    def create_client(self) -> MongoClient:
        """Returns a client for a MongoDB instance"""
        uri = (self.configuration.secrets or {}).get("url") or self.build_uri()
        try:
            return MongoClient(
                uri, serverSelectionTimeoutMS=5000, uuidRepresentation="standard"
            )
        except ValueError:
            raise ConnectionException("Value Error connecting to MongoDB.")

    def query_config(self, node: ExecutionNode) -> QueryConfig[Any]:
        """Query wrapper corresponding to the input traversal_node."""
        return MongoQueryConfig(node)

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Connects to the Mongo database and makes two trivial queries to ensure connection is valid.
        """
        logger.info("Starting test connection to {}", self.configuration.key)
        client = self.client()

        try:
            # Make a couple of trivial requests - getting server info and fetching the collection names
            client.server_info()

            # Retrieve the default auth database if it exists
            default_auth_db = (
                self.configuration.secrets.get("defaultauthdb")
                if self.configuration.secrets
                else None
            )
            if default_auth_db:
                db = client[default_auth_db]
                db.list_collection_names()
        except ServerSelectionTimeoutError:
            raise ConnectionException(
                "Server Selection Timeout Error connecting to MongoDB."
            )
        except OperationFailure:
            raise ConnectionException("Operation Failure connecting to MongoDB.")
        except Exception:
            raise ConnectionException("Connection Error connecting to MongoDB.")
        finally:
            client.close()

        return ConnectionTestStatus.succeeded

    def retrieve_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """Retrieve mongo data"""
        query_config = self.query_config(node)
        client = self.client()

        query_components = query_config.generate_query(input_data, policy)
        if query_components is None:
            return []
        query_data, fields = query_components

        db_name = node.address.dataset
        collection_name = node.address.collection

        db = client[db_name]
        collection = db[collection_name]
        rows = []
        logger.info("Starting data retrieval for {}", node.address)
        for row in collection.find(query_data, fields):
            rows.append(row)
        logger.info("Found {} rows on {}", len(rows), node.address)
        return rows

    def mask_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        rows: List[Row],
    ) -> int:
        """Execute a masking request"""
        query_config = self.query_config(node)
        collection_name = node.address.collection
        client = self.client()
        update_ct = 0
        for row in rows:
            update_stmt = query_config.generate_update_stmt(
                row, policy, privacy_request
            )
            if update_stmt is not None:
                query, update = update_stmt
                db = client[node.address.dataset]
                collection = db[collection_name]
                update_result = collection.update_one(query, update, upsert=False)
                update_ct += update_result.modified_count
                logger.info(
                    "db.{}.update_one({}, {}, upsert=False)",
                    collection_name,
                    Pii(query),
                    Pii(update),
                )

        return update_ct

    def close(self) -> None:
        """Close any held resources"""
        if self.db_client:
            self.db_client.close()



"""migrate translation data - notices and experiences

Revision ID: 14acee6f5459
Revises: 0c65325843bd
Create Date: 2024-01-09 21:17:13.115020

"""

import csv
import uuid
from datetime import datetime
from enum import Enum
from typing import Optional

from alembic import op
from loguru import logger
from sqlalchemy import text
from sqlalchemy.engine import ResultProxy

from fides.api.alembic.migrations.helpers.database_functions import generate_record_id

# revision identifiers, used by Alembic.

revision = "14acee6f5459"
down_revision = "0c65325843bd"
branch_labels = None
depends_on = None


EXPERIENCE_CONFIG_DEFAULTS = [
    {
        "id": "pri-76b8-cc52-11ee-9eaa-8ef97a-modal-config",
        "name": "US Modal",
        "regions": ["us_ca", "us_co", "us_ct", "us_ut", "us_va"],
        "component": "modal",
        "allow_language_selection": False,
        "auto_detect_language": True,
        "dismissable": True,
        "disabled": True,
        "privacy_notice_keys": ["data_sales_and_sharing"],
        "origin": "pri-76b8-cc52-11ee-9eaa-8ef97a-modal",
    },
    {
        "id": "pri-27d4-cc53-11ee-9eaa-8ef97a04-pri-config",
        "name": "US Privacy Center",
        "regions": ["us_ca", "us_co", "us_ct", "us_ut", "us_va"],
        "component": "privacy_center",
        "allow_language_selection": False,
        "auto_detect_language": True,
        "dismissable": True,
        "disabled": True,
        "privacy_notice_keys": ["data_sales_and_sharing"],
        "origin": "pri-27d4-cc53-11ee-9eaa-8ef97a04-pri",
    },
    {
        "id": "pri-c8ff-78d6-4a02-850f-2c0ban-modal-config",
        "name": "EEA Banner + Modal",
        "regions": [
            "be",
            "bg",
            "cz",
            "dk",
            "de",
            "ee",
            "ie",
            "gr",
            "es",
            "fr",
            "hr",
            "it",
            "cy",
            "lv",
            "lt",
            "lu",
            "hu",
            "mt",
            "nl",
            "at",
            "pl",
            "pt",
            "ro",
            "si",
            "sk",
            "fi",
            "se",
            "gb",
            "is",
            "no",
            "li",
        ],
        "component": "banner_and_modal",
        "allow_language_selection": False,
        "auto_detect_language": True,
        "dismissable": True,
        "disabled": True,
        "privacy_notice_keys": ["essential", "functional", "analytics", "marketing"],
        "origin": "pri-c8ff-78d6-4a02-850f-2c0ban-modal",
    },
    {
        "id": "pri-694e-02bd-4afe-81b3-2a2ban-modal-config",
        "name": "Canada Banner + Modal",
        "regions": [
            "ca",
        ],
        "component": "banner_and_modal",
        "allow_language_selection": False,
        "auto_detect_language": True,
        "dismissable": True,
        "disabled": True,
        "privacy_notice_keys": ["marketing"],
        "origin": "pri-694e-02bd-4afe-81b3-2a2ban-modal",
    },
    {
        "id": "pri-d9ed6-25e7-4953-8977-772ban-modal-config",
        "name": "Quebec Banner + Modal",
        "regions": [
            "ca_qc",
        ],
        "component": "banner_and_modal",
        "allow_language_selection": False,
        "auto_detect_language": True,
        "dismissable": True,
        "disabled": True,
        "privacy_notice_keys": ["essential", "analytics", "marketing"],
        "origin": "pri-d9ed6-25e7-4953-8977-772ban-modal",
    },
    {
        "id": "d7c3ce0a-a3b3-43ff-bf6f-3d8-tcf-over-config",
        "name": "TCF",
        "allow_language_selection": False,
        "auto_detect_language": True,
        "dismissable": True,
        "disabled": True,
        "notices": [],
        "regions": [
            "be",
            "bg",
            "cz",
            "dk",
            "de",
            "ee",
            "ie",
            "gr",
            "es",
            "fr",
            "hr",
            "it",
            "cy",
            "lv",
            "lt",
            "lu",
            "hu",
            "mt",
            "nl",
            "at",
            "pl",
            "pt",
            "ro",
            "si",
            "sk",
            "fi",
            "se",
            "gb",
            "is",
            "no",
            "li",
            "eea",
        ],
        "component": "tcf_overlay",
        "privacy_notice_keys": [],
        "origin": "d7c3ce0a-a3b3-43ff-bf6f-3d8-tcf-over",
    },
]

# the known existing (legacy) experience config IDs.
# we will only attempt to migrate config data from these experience configs
KNOWN_EXISTING_EXPERIENCE_CONFIGS = {
    "pri-7ae3-f06b-4096-970f-0bbbdef-over",
    "pri-097a-d00d-40b6-a08f-f8e50def-pri",
    "a4974670-abad-471f-9084-2cb-tcf-over",
}


# this ties a DB record ID to our logical identifier of the type of OOB experience.
# the DB record ID is based on what we've defined in the associated config yml.
class DefaultExperienceConfigTypes(Enum):
    EEA_BANNER_AND_MODAL = "pri-c8ff-78d6-4a02-850f-2c0ban-modal-config"
    EEA_TCF_OVERLAY = "d7c3ce0a-a3b3-43ff-bf6f-3d8-tcf-over-config"
    US_MODAL = "pri-76b8-cc52-11ee-9eaa-8ef97a-modal-config"
    US_PRIVACY_CENTER = "pri-27d4-cc53-11ee-9eaa-8ef97a04-pri-config"
    CANADA_BANNER_MODAL = "pri-694e-02bd-4afe-81b3-2a2ban-modal-config"
    QUEBEC_BANNER_MODAL = "pri-d9ed6-25e7-4953-8977-772ban-modal-config"


class ComponentType(Enum):
    PrivacyCenter = "privacy_center"
    Overlay = "overlay"
    TCFOverlay = "tcf_overlay"


# our new OOB EEA (EU) experience configs, keyed by their component type
EEA_MAPPING = {
    ComponentType.Overlay: DefaultExperienceConfigTypes.EEA_BANNER_AND_MODAL,
    ComponentType.TCFOverlay: DefaultExperienceConfigTypes.EEA_TCF_OVERLAY,
}

# our new OOB US experience configs, keyed by their component type
US_MAPPING = {
    ComponentType.PrivacyCenter: DefaultExperienceConfigTypes.US_PRIVACY_CENTER,
    ComponentType.Overlay: DefaultExperienceConfigTypes.US_MODAL,
}

# our new OOB Canada experience configs, keyed by their component type
CANADA_MAPPING = {
    ComponentType.Overlay: DefaultExperienceConfigTypes.CANADA_BANNER_MODAL,
}

QUEBEC_MAPPING = {
    ComponentType.Overlay: DefaultExperienceConfigTypes.QUEBEC_BANNER_MODAL,
}

# this map contains the biggest _assumption_ about the migration -
# these are the notices that will be associated with our new OOB experience configs.
# any existing notices besides these will remain "orphaned" post-migration.
# additionally, the regions that (pre-migration) are associated with these notices
# will be used to infer the regions associated with the post-migration experience
# configs to which they map.
NOTICES_TO_EXPERIENCE_CONFIG = {
    "marketing": [EEA_MAPPING, CANADA_MAPPING, QUEBEC_MAPPING],
    "functional": [EEA_MAPPING],
    "essential": [EEA_MAPPING, QUEBEC_MAPPING],
    "analytics": [EEA_MAPPING, QUEBEC_MAPPING],
    "data_sales_and_sharing": [US_MAPPING],
}


# as part of this upgrade, we're removing the following GB sub-regions as region options across the app
GB_REGIONS_TO_REMOVE = {
    "gb_eng",
    "gb_sct",
    "gb_wls",
    "gb_nir",
}
# the GB sub-regions are replaced by the `gb` region
GB_REGIONS_TO_ADD = {
    "gb",
}

# Canadian regions are ignored in the migration of existing data, since we create a net-new OOB experience config for Canada
CA_REGIONS_TO_IGNORE = {
    "ca_qc",
    "ca_nt",
    "ca_nb",
    "ca_nu",
    "ca_bc",
    "ca_mb",
    "ca_ns",
    "ca_pe",
    "ca_sk",
    "ca_yt",
    "ca_on",
    "ca_nl",
    "ca_ab",
    "ca",
}


def dump_table_to_csv(bind, table_name):
    csv_file_path = f'./multilang_migration_backup_{table_name}_{datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}.csv'
    query = text(f"SELECT * FROM {table_name}")
    result = bind.execute(query)

    # Get column names from the result set
    column_names = result.keys()

    try:
        # Write results to CSV
        with open(csv_file_path, "w") as csv_file:
            csv_writer = csv.writer(csv_file)

            # Write header row
            csv_writer.writerow(column_names)

            # Write data rows
            for row in result:
                csv_writer.writerow(row)
    except PermissionError:
        logger.exception(
            f"Permission error writing migration backup file {csv_file_path}"
        )


def load_default_experience_configs():
    """
    Loads default experience config definitions from yml file.

    Returns two dicts in a tuple:
        - first dict holds "new" experience config records that will be used to reconcile existing data
        - second dict holds the "raw" experience config records based on the definitions on disk. these will be used as a reference for e.g. template creation.
    """
    _raw_experience_config_map = (
        {}
    )  # this will just hold config templates exactly as they are in the 'oob' file
    _reconciled_experience_config_map = (
        {}
    )  # will hold experience configs 'reconciled' between existing data and 'oob' template data

    for experience_config in EXPERIENCE_CONFIG_DEFAULTS:
        experience_config_type = DefaultExperienceConfigTypes(experience_config["id"])
        _raw_experience_config_map[experience_config_type] = experience_config.copy()

        # the reconciled experience config will have its regions populated by existing notices that are
        # associated with this experience config. so in most cases, we start "fresh" with an empty set.
        # the TCF experience and Canada/Quebec banner modal experiences are exceptions:
        # their regions are not derived from existing notices (TCF is not linked to notices; Canada experience config is net-new),
        # so we hardcode their regions to the config defaults.
        if experience_config_type not in (
            DefaultExperienceConfigTypes.EEA_TCF_OVERLAY,
            DefaultExperienceConfigTypes.CANADA_BANNER_MODAL,
            DefaultExperienceConfigTypes.QUEBEC_BANNER_MODAL,
        ):
            experience_config["regions"] = set()
        else:
            experience_config["regions"] = set(experience_config["regions"])

        experience_config["privacy_notices"] = set()
        experience_config["needs_migration"] = (
            False  # indicator whether the record requires migration, or we can default to the template values
        )
        _reconciled_experience_config_map[experience_config_type] = experience_config
        experience_config["needs_migration"] = False
    return _reconciled_experience_config_map, _raw_experience_config_map


(
    reconciled_experience_config_map,
    raw_experience_config_map,
) = load_default_experience_configs()


def backup_existing_data_csv(bind):
    """
    Perform a backup (dump to csv) of the data in our tables that will be impacted by migration.

    Because this data migration is risky, it is recommended to perform a comprehensive DB backup/snapshot beforehand!
    The "automated" dump to CSV here is an additional safety measure.
    """
    for table_name in [
        "privacyexperience",
        "privacyexperienceconfig",
        "privacyexperienceconfighistory",
        "privacynotice",
        "privacynoticehistory",
    ]:
        try:
            dump_table_to_csv(bind, table_name)
        except Exception:
            # permission error is handled within the function, but just to be extra safe
            # let's make sure our migration proceeds if ANY error is thrown from the CSV dump,
            # since this is just a nice-to-have.
            logger.exception(f"Error backing up table {table_name} to disk")


def remove_existing_experience_data(bind):
    """
    We remove any existing PrivacyExperience and PrivacyExperienceConfig records,
    to allow for a "clean-slate" in creation of _new_ PrivacyExperienceConfigs based on the
    "out of the box" templates.

    This means that any existing user-edited data on PrivacyExperienceConfig records will need to be
    _manually_ re-added to the post-migration state. It will not be migrated as part of the automatic migration.
    """

    def delete_experiences(bind):
        """
        Delete all PrivacyExperience records.

        These records will be effectively re-created as a result of the "out of the box" creation
        of 'new-style' PrivacyExperienceConfig records post-migration.
        """
        bind.execute(
            text(
                """
                DELETE FROM privacyexperience;
                """
            )
        )

    def delete_experience_configs(bind):
        """
        Delete all PrivacyExperienceConfig records.

        These records will be effectively re-created via the server's "out of the box" creation
        of the updated PrivacyExperienceConfig records, post-migration.
        """
        bind.execute(
            text(
                """
                DELETE FROM privacyexperienceconfig;
                """
            )
        )

    delete_experiences(bind)
    delete_experience_configs(bind)


def is_canada_or_quebec_experience(
    config_type: Optional[DefaultExperienceConfigTypes],
) -> bool:
    """Helper to determine whether the target experience config type is a canada or quebec experience"""
    return config_type in {
        DefaultExperienceConfigTypes.CANADA_BANNER_MODAL,
        DefaultExperienceConfigTypes.QUEBEC_BANNER_MODAL,
    }


def determine_needed_experience_configs(bind):
    """
    Our data model changes necessitate the creation of new Experience Configs. Now that "where" a Notice is served
    is being moved from the "Notice" constructs to "Experiences", existing Experience Configs are not sufficient
    to capture the full range of configuration on current Notices.
    """

    notice_query = """select id, name, regions, disabled, displayed_in_overlay, displayed_in_privacy_center, notice_key from privacynotice;"""
    experience_config_query = """SELECT id, component, disabled, accept_button_label, description, privacy_preferences_link_label, privacy_policy_link_label, privacy_policy_url, save_button_label, title, banner_description, banner_title, reject_button_label, acknowledge_button_label from privacyexperienceconfig;"""

    existing_notices: ResultProxy = bind.execute(text(notice_query))
    for existing_notice in existing_notices:
        notice_key = existing_notice["notice_key"]
        component_mappings = NOTICES_TO_EXPERIENCE_CONFIG.get(notice_key, [])
        if component_mappings:
            for component_mapping in component_mappings:
                if is_canada_or_quebec_experience(
                    component_mapping.get(ComponentType.Overlay)
                ):
                    # If this is a Canadian/Quebec Banner and Modal, skip updating regions, and just use OOB
                    regions_to_add = set()
                else:
                    regions_to_add = set(existing_notice["regions"]).copy()

                    # remove GB subregions and replace with top-level `gb` region!
                    if not GB_REGIONS_TO_REMOVE.isdisjoint(
                        regions_to_add
                    ):  # if we have GB regions to remove
                        regions_to_add.difference_update(
                            GB_REGIONS_TO_REMOVE
                        )  # remove the GB regions
                        regions_to_add.update(GB_REGIONS_TO_ADD)  # add the GB region

                    # ignore any canada regions in migration - they are not migrated,
                    # but instead handled by a net-new OOB experience
                    regions_to_add.difference_update(CA_REGIONS_TO_IGNORE)

                if existing_notice["displayed_in_overlay"]:
                    experience_config = component_mapping.get(ComponentType.Overlay)
                    if experience_config:
                        reconciled_experience_config_map[experience_config][
                            "regions"
                        ].update(regions_to_add)
                        reconciled_experience_config_map[experience_config][
                            "privacy_notices"
                        ].add(existing_notice["id"])
                        if not existing_notice[
                            "disabled"
                        ] and not is_canada_or_quebec_experience(experience_config):
                            # Disabled by default, but if not a Canada/Quebec experience and any notices are enabled,
                            # let's enable this Experience Config as the safer default
                            reconciled_experience_config_map[experience_config][
                                "disabled"
                            ] = False
                        reconciled_experience_config_map[experience_config][
                            "needs_migration"
                        ] = True

                if existing_notice["displayed_in_privacy_center"]:
                    experience_config = component_mapping.get(
                        ComponentType.PrivacyCenter
                    )
                    if experience_config:
                        reconciled_experience_config_map[experience_config][
                            "regions"
                        ].update(regions_to_add)
                        reconciled_experience_config_map[experience_config][
                            "privacy_notices"
                        ].add(existing_notice["id"])
                        if not existing_notice[
                            "disabled"
                        ] and not is_canada_or_quebec_experience(experience_config):
                            # Disabled by default, but if not a Canada/Quebec experience and any notices are enabled,
                            # let's enable this Experience Config as the safer default
                            reconciled_experience_config_map[experience_config][
                                "disabled"
                            ] = False

                        reconciled_experience_config_map[experience_config][
                            "needs_migration"
                        ] = True

    existing_experience_configs: ResultProxy = bind.execute(
        text(experience_config_query)
    )
    has_existing_experience_configs = False
    for eec in existing_experience_configs:
        # only migrate experience config data if its a known config
        if eec["id"] in KNOWN_EXISTING_EXPERIENCE_CONFIGS:
            has_existing_experience_configs = True
            component = ComponentType(eec["component"])
            configs = []

            # because we've given experiences a location dimension in this migration,
            # an existing experience effectively maps to two "new" default experiences -
            # one that is surfaced in the US (and is associated with "US" notices), and
            # one that is surfaced in the EU (and is associated with the "EU" notices).
            # because of this, we'll copy over existing experience text/attributes to _both_
            # of these corresponding new experience configs.
            if eea_config_id := EEA_MAPPING.get(component):
                configs.append(reconciled_experience_config_map.get(eea_config_id))
            if us_config_id := US_MAPPING.get(component):
                configs.append(reconciled_experience_config_map.get(us_config_id))
            if canada_config_id := CANADA_MAPPING.get(component):
                configs.append(reconciled_experience_config_map.get(canada_config_id))
            if quebec_config_id := QUEBEC_MAPPING.get(component):
                configs.append(reconciled_experience_config_map.get(quebec_config_id))
            for config in configs:
                if config:
                    config["accept_button_label"] = eec["accept_button_label"]
                    config["acknowledge_button_label"] = eec["acknowledge_button_label"]
                    config["banner_description"] = eec["banner_description"]
                    config["banner_title"] = eec["banner_title"]
                    config["description"] = eec["description"]
                    config["privacy_policy_link_label"] = eec[
                        "privacy_policy_link_label"
                    ]
                    config["privacy_policy_url"] = eec["privacy_policy_url"]
                    config["privacy_preferences_link_label"] = eec[
                        "privacy_preferences_link_label"
                    ]
                    config["reject_button_label"] = eec["reject_button_label"]
                    config["save_button_label"] = eec["save_button_label"]
                    config["title"] = eec["title"]
                    config["needs_migration"] = True

                    if eec["component"] == "tcf_overlay":
                        # TCF doesn't have notices so enable the new TCF overlay if its existing overlay is enabled
                        config["disabled"] = eec["disabled"]

    if has_existing_experience_configs:
        return reconciled_experience_config_map

    # if no existing experience configs are found, don't return the map, because no experience config migration is needed.
    # instead, we'll rely on the OOB loading of experience configs on server startup.
    # basically, this is allows us to bypass the data migration if we're not actually migrating a db that's been in use!
    return {}


def migrate_experiences(bind):
    """
    Migrates data from any existing ExperienceConfig records to the new, translation-based data model, and reconciles
    the old ExperienceConfig records with new, OOB ExperienceConfig records.

    - First the necessary existing ExperienceConfig and PrivacyNotice data is retrieved and stored in memory
    - It is then, in memory, reconciled with our new OOB ExperienceConfig records, which are loaded from a YAML definition on disk
    - Then, we remove all existing Experience and ExperienceConfig records to start from a 'blank slate'
    - Then, based on the reconciled existing ExperienceConfig data, we create new ExperienceConfig and Experience records
    """

    def create_new_experience_config_templates(experience_configs):
        for experience_config in experience_configs:
            # for config template record:
            # - `id` is the `origin` field from the config template record loaded from disk
            create_experience_config_template_query = text(
                """
                INSERT INTO experienceconfigtemplate (id, component, name, disabled, allow_language_selection, dismissable, privacy_notice_keys, regions, auto_detect_language)
                VALUES (:id, :component, :name, :disabled, :allow_language_selection, :dismissable, :privacy_notice_keys, :regions, :auto_detect_language)
                """
            )

            bind.execute(
                create_experience_config_template_query,
                {
                    **experience_config,
                    "id": experience_config[
                        "origin"
                    ],  # origin of the config record is the template ID
                },
            )

    def create_new_experience_config(experience_config):
        """
        Creates a new experience config based on the provided definition. Also creates corresponding history record.

        """
        # for config record:
        # - `id` is the `id` grabbed from our config template record
        # - `origin` is the `origin` grabbed from the config template
        create_experience_config_query = text(
            """
            INSERT INTO privacyexperienceconfig (id, version, origin, component, name, disabled, is_default, allow_language_selection, dismissable, auto_detect_language, accept_button_label, description, privacy_preferences_link_label, privacy_policy_link_label, privacy_policy_url, save_button_label, title, banner_description, banner_title, reject_button_label, acknowledge_button_label)
            VALUES (:id,  :version, :origin, :component, :name, :disabled, :is_default, :allow_language_selection, :dismissable, :auto_detect_language, :accept_button_label, :description, :privacy_preferences_link_label, :privacy_policy_link_label, :privacy_policy_url, :save_button_label, :title, :banner_description, :banner_title, :reject_button_label, :acknowledge_button_label)
            """
        )

        bind.execute(
            create_experience_config_query,
            {
                **experience_config,
                "version": 1,
                "is_default": True,
            },
        )

        # for history record:
        # - `id` is a generated UUID
        # - `experience_config_id` is the `id` grabbed from our config template record on disk
        # - `origin` is the `origin` grabbed from the config template record on disk
        create_experience_config_history = text(
            """
             INSERT INTO privacyexperienceconfighistory (id, experience_config_id, origin, name, disabled, version, component, is_default, allow_language_selection, auto_detect_language, dismissable, accept_button_label, description, privacy_preferences_link_label, privacy_policy_link_label, privacy_policy_url, save_button_label, title, banner_description, banner_title, reject_button_label, acknowledge_button_label, language)
             VALUES (:history_id, :id, :origin, :name, :disabled, :version, :component, :is_default, :allow_language_selection, :auto_detect_language, :dismissable, :accept_button_label, :description, :privacy_preferences_link_label, :privacy_policy_link_label, :privacy_policy_url, :save_button_label, :title, :banner_description, :banner_title, :reject_button_label, :acknowledge_button_label, :language)
            """
        )

        bind.execute(
            create_experience_config_history,
            {
                **experience_config,
                "history_id": generate_record_id("pri"),
                "version": 1,
                "is_default": True,
                "language": "en",
            },
        )

    def create_applicable_experiences(experience_config):
        """
        Creates all the applicable experience records based on the provided experience config definition.

        i.e., looks at all regions on the provided experience config definition, and creates a corresponding
        experience record for each region.
        """
        new_experience = text(
            """
            INSERT INTO privacyexperience (id, experience_config_id, region)
            VALUES(:experience_id, :experience_config_id, :region)
            """
        )
        for region in experience_config["regions"]:
            if region is not None:
                bind.execute(
                    new_experience,
                    {
                        "experience_id": generate_record_id("pri"),
                        "experience_config_id": experience_config["id"],
                        "region": region,
                    },
                )

    def link_notices_to_experiences(experience_config):
        """
        Creates experience notice records based on the provided experience config definition.

        i.e., looks at all `privacy_notices` linked to the experience config, and creates a corresponding
        experience notice record for each linked notice.
        """
        link_notice_query = text(
            """
                INSERT INTO experiencenotices (id, notice_id, experience_config_id)
                VALUES (:record_id, :notice_id, :experience_config_id)
                """
        )
        for notice_id in experience_config["privacy_notices"]:
            bind.execute(
                link_notice_query,
                {
                    "record_id": generate_record_id("exp"),
                    "notice_id": notice_id,
                    "experience_config_id": experience_config["id"],
                },
            )

    def create_experience_translations(bind):
        """
        Creates new experience translation records for each privacy experience config record in the db.

        We simply take the text that's been temporarily saved on the config record directly and move it
        over to the translation record. The translation is assumed to be english.

        This requires the privacy experience config records to have been created before this is invoked!

        """
        get_experience_config_ids = text(
            """SELECT id FROM privacyexperienceconfig;
         """
        )

        # for translation record:
        # - `id` is a generated UUID
        # - `experience_config_id` is the `id` column grabbed from the experience config query result
        # - language is assumed to be english!
        create_translation_query = text(
            """
            INSERT INTO experiencetranslation (id, experience_config_id, created_at, updated_at, language, is_default, accept_button_label, reject_button_label, acknowledge_button_label, banner_description, banner_title, description, title,  privacy_preferences_link_label, privacy_policy_link_label, privacy_policy_url, save_button_label)
            SELECT :record_id, id, created_at, updated_at, :language, :is_default, accept_button_label, reject_button_label, acknowledge_button_label, banner_description, banner_title, description, title,  privacy_preferences_link_label, privacy_policy_link_label, privacy_policy_url, save_button_label
            FROM privacyexperienceconfig
            WHERE id = :experience_config_id
            """
        )
        for res in bind.execute(get_experience_config_ids):
            # Create a translation for each Experience
            bind.execute(
                create_translation_query,
                {
                    "record_id": generate_record_id("exp"),
                    "language": "en",
                    "experience_config_id": res["id"],
                    "is_default": True,
                },
            )

        # Repoint the Experience Config History to the Experience Translation
        link_config_to_translation_id = text(
            """
            UPDATE privacyexperienceconfighistory
            SET translation_id = experiencetranslation.id
            FROM experiencetranslation
            WHERE experiencetranslation.experience_config_id = privacyexperienceconfighistory.experience_config_id;
        """
        )

        bind.execute(link_config_to_translation_id)

    # first extract existing experience (config) data needed for migration
    experience_configs = determine_needed_experience_configs(bind)

    # before wiping data, let's perform a csv backup to disk just to be extra safe
    backup_existing_data_csv(bind)

    # wipe the existing experience + experience config data to start from a blank slate
    remove_existing_experience_data(bind)

    # if we don't have experience configs, we can bypass all experience config data migration
    if experience_configs:
        # first create experience config templates based on raw experience config definitions
        create_new_experience_config_templates(raw_experience_config_map.values())

        # create new experience + experience configs based on old experience config data
        # these have been reconciled with the new OOB experience config records
        for experience_config_type in DefaultExperienceConfigTypes:
            logger.info(f"Migrating experience config {experience_config_type.name}")
            reconciled_experience_config = experience_configs.get(
                experience_config_type, {}
            )
            raw_experience_config = raw_experience_config_map.get(
                experience_config_type, {}
            )

            if not reconciled_experience_config and not raw_experience_config:
                logger.info(
                    f"No reconciled or raw experience config found for {experience_config_type.name}"
                )
                continue

            # revert to the "raw" config if we haven't flagged a config as needing migration
            # in practice, we should always have a reconciled record here that's flagged for migration,
            # but this covers our bases in case anyone has e.g. deleted a previous OOB notice.
            if reconciled_experience_config.get("needs_migration", False):
                logger.info(
                    f"Migrating reconciled experience config {reconciled_experience_config['id']}"
                )
                experience_config_type = reconciled_experience_config
            else:
                logger.info(
                    f"Creating raw experience config {raw_experience_config['id']}"
                )
                experience_config_type = raw_experience_config

            create_new_experience_config(experience_config_type)
            create_applicable_experiences(experience_config_type)
            link_notices_to_experiences(experience_config_type)

        # experience translations are then created based on the new experience config records
        create_experience_translations(bind)


def migrate_notices(bind):
    """
    Migrates existing PrivacyNotice records to the new, translation-based data model.

    Notice keys are not impacted. Most existing notice data is moved onto a linked NoticeTranslation record,
    which is assumed to be english-language as a default. If the existing notice text/content is _not_ english-language,
    then users will need to *manually* adjust the language on the created translation record.

    The existing PrivacyNoticeHistory records are adjusted to point to the associated NoticeTranslation record,
    as is expected in the new data model.
    """

    get_notices_query = text(
        """SELECT id from privacynotice;
         """
    )

    create_notice_translation_query = text(
        """
        INSERT INTO noticetranslation (id, privacy_notice_id, language, title, description, created_at, updated_at)
        SELECT :record_id, :original_notice_id, :language, name, description, created_at, updated_at
        FROM privacynotice
        WHERE id = :original_notice_id
        """
    )
    for res in bind.execute(get_notices_query):
        # Create a Notice Translation for each notice, setting to english as a default
        bind.execute(
            create_notice_translation_query,
            {
                "record_id": generate_record_id("pri"),
                "language": "en",
                "original_notice_id": res["id"],
            },
        )

    # Add a FK from the Notice History back to the Notice Translation
    link_notice_history_to_translation_id = text(
        """
        UPDATE privacynoticehistory
        SET translation_id = noticetranslation.id, title = privacynoticehistory.name
        FROM noticetranslation
        WHERE noticetranslation.privacy_notice_id = privacynoticehistory.privacy_notice_id;
        """
    )

    bind.execute(link_notice_history_to_translation_id)

    # Queries the latest historical version for each notice
    new_notice_histories_to_create_query = text(
        """
        select privacy_notice_id, max(version) from privacynoticehistory group by privacy_notice_id;
        """
    )
    # Create a new privacy notice history record that bumps the version and clears regions and displayed_in* fields and sets language
    for res in bind.execute(new_notice_histories_to_create_query):
        create_privacy_notice_history_query = text(
            """
            INSERT INTO privacynoticehistory (id, name, description, origin, consent_mechanism, data_uses, version, disabled, enforcement_level, has_gpc_flag, internal_description, notice_key, gpp_field_mapping, framework, language, title, translation_id, privacy_notice_id)
            SELECT :record_id, name, description, origin, consent_mechanism, data_uses, :new_version, disabled, enforcement_level, has_gpc_flag, internal_description, notice_key, gpp_field_mapping, framework, :language, title, translation_id, privacy_notice_id
            FROM privacynoticehistory
            WHERE version = :current_version AND
            privacy_notice_id = :privacy_notice_id
            ORDER BY created_at DESC LIMIT 1
        """
        )
        bind.execute(
            create_privacy_notice_history_query,
            {
                "record_id": generate_record_id("pri"),
                "language": "en",
                "new_version": res["max"] + 1,
                "current_version": res["max"],
                "privacy_notice_id": res["privacy_notice_id"],
            },
        )


def downward_migrate_notices(bind):
    """
    Back-migration of PrivacyNotice data.

    Moves data in existing NoticeTranslation records back to the associated PrivacyNotice records,
    to match the old data model.
    # NOTE: this won't work _well_ if multiple translations are present. Should we even attempt it?

    The existing PrivacyNoticeHistory records are adjusted to point back to the associated PrivacyNotice records,
    rather than the NoticeTranslation records, as is expected in the old data model.

    """
    noticetranslation_data_to_notice_query = text(
        """
        UPDATE privacynotice
        SET name = noticetranslation.title, description = noticetranslation.description
        FROM noticetranslation
        WHERE noticetranslation.privacy_notice_id = privacynotice.id
        """
    )
    bind.execute(noticetranslation_data_to_notice_query)

    # Update the FK from the Notice History back to the Notice record
    link_notice_history_to_notice = text(
        """
        UPDATE privacynoticehistory
        SET privacy_notice_id = noticetranslation.privacy_notice_id
        FROM noticetranslation
        WHERE noticetranslation.id = privacynoticehistory.translation_id;
    """
    )

    bind.execute(link_notice_history_to_notice)


def upgrade():
    logger.info(
        f"Starting Consent Multitranslation Data Migration (#2) {datetime.now()}"
    )

    bind = op.get_bind()

    migrate_experiences(bind)

    migrate_notices(bind)


def downgrade():
    # Downgrade will _not_ attempt to recreate PrivacyExperienceConfig recrords,
    # as that should be handled by the (old) server's OOB loading behavior
    bind = op.get_bind()

    downward_migrate_notices(bind)

    bind.execute(
        text(
            """
            DELETE FROM experiencetranslation;
            """
        )
    )
    remove_existing_experience_data(
        bind
    )  # remove existing experience data to ensure it won't prevent startup on downgraded versions!



"""data_mig_tcf_purpose_header_and_descript

Data migration to add new TCF purpose headers and update
existing TCF descriptions, provided they haven't been updated yet.

Revision ID: cc37edf20859
Revises: 1332815dcd71
Create Date: 2024-08-29 20:43:56.657171

"""

from datetime import datetime

from alembic import op
from loguru import logger
from sqlalchemy import text

from fides.api.alembic.migrations.helpers.database_functions import generate_record_id

# revision identifiers, used by Alembic.
revision = "cc37edf20859"
down_revision = "1332815dcd71"
branch_labels = None
depends_on = None

# fmt: off
updated_translations = {
    "lv": {
        "purpose_header": "Mēs kopā ar partneriem apstrādājam datus šādiem mērķiem",
        "description": "Mēs un mūsu __VENDOR_COUNT_LINK__ partneri izmantojam sīkfailus un citas saistītās tehnoloģijas, lai palīdzētu mums identificēt apmeklētājus un saglabāt viņu preferences. Šīs tehnoloģijas varam izmantot arī, lai mērķētu reklāmas, izmērītu reklāmas kampaņu panākumu līmeni un pārbaudītu vietnes trafiku. Lai gan dažas no šīm tehnoloģijām nav obligātas, tomēr tās palīdz uzlabot lietotāja pieredzi dažādos veidos, citas savukārt ir nepieciešamas, lai garantētu, ka pakalpojums vai vietne darbojas, atbilstošā veidā, un tās nedrīkst atspējot. \n\nLai apstrādātu personas datus, mēs kopā ar __VENDOR_COUNT_LINK__ partneriem saglabājam un/vai piekļūstam informācijai lietotāja ierīcē, piemēram, IP adresēm, unikālajiem identifikatoriem un sīkfailos saglabātajiem pārlūkošanas datiem. Izvēloties opciju \"Pārvaldīt iestatījumus\", kas atrodama lapas kājenē, varat kontrolēt savus iestatījumus. Lūdzu, apmeklējiet mūsu pārdevēju/ pakalpojumu sniedzēju lapu, lai pārbaudītu vai iebilstu pret jebkādiem gadījumiem, kad mūsu partneri apgalvo, ka viņiem ir leģitīmais pamats izmantot jūsu datus."
    },
    "et": {
        "purpose_header": "Meie ja meie partnerid töötleme andmeid järgmistel eesmärkidel",
        "description": "Meie ja meie __VENDOR_COUNT_LINK__ partnerid kasutame küpsiseid ja sarnaseid meetodeid külastajate äratundmiseks ja nende eelistuste meelespidamiseks. Võime neid tehnoloogiaid kasutada ka reklaamikampaaniate tõhususe hindamiseks, reklaami suunamiseks ja veebisaidi külastatavuse analüüsimiseks. Mõned neist tehnoloogiatest on teenuse või veebisaidi nõuetekohase toimimise tagamiseks hädavajalikud ja neid ei saa keelata, teised on vabatahtlikud, kuid parandavad eri viisidel kasutajakogemust. \n\nKoostöös oma __VENDOR_COUNT_LINK__ partneritega salvestame ja/või kasutame kasutaja seadmes olevat teavet, sealhulgas, kuid mitte ainult, IP-aadresse, unikaalseid identifikaatoreid ja küpsistesse salvestatud sirvimisandmeid, et töödelda isikuandmeid. Teil on võimalus oma eelistusi hallata, valides lehekülje jaluses oleva valiku „Eelistuste haldamine“. Kui soovite vaadata juhtumeid, kus meie partnerid väidavad, et neil on õigustatud huvi teie andmete kasutamiseks, või neile juhtumitele vastuväiteid esitada, külastage palun meie kaupmeeste lehekülge."
    },
    "el": {
        "purpose_header": "Εμείς και οι συνεργάτες μας επεξεργαζόμαστε δεδομένα για τους ακόλουθους σκοπούς",
        "description": "Εμείς και οι συνεργάτες μας __VENDOR_COUNT_LINK__ χρησιμοποιούμε cookie και παρόμοιες μεθόδους για να αναγνωρίζουμε τους επισκέπτες και να θυμόμαστε τις προτιμήσεις τους. Ενδέχεται επίσης να χρησιμοποιήσουμε αυτές τις τεχνολογίες για να μετρήσουμε την αποτελεσματικότητα των διαφημιστικών εκστρατειών, να στοχεύσουμε διαφημίσεις και να αναλύσουμε την επισκεψιμότητα του ιστότοπου. Ορισμένες από αυτές τις τεχνολογίες είναι απαραίτητες για τη διασφάλιση της σωστής λειτουργίας της υπηρεσίας ή του ιστότοπου και δεν μπορούν να απενεργοποιηθούν, ενώ άλλες είναι προαιρετικές, αλλά χρησιμεύουν για τη βελτίωση της εμπειρίας του χρήστη με διάφορους τρόπους. \n\nΕμείς, σε συνεργασία με τους συνεργάτες μας __VENDOR_COUNT_LINK__, αποθηκεύουμε ή/και έχουμε πρόσβαση σε πληροφορίες στη συσκευή ενός χρήστη, συμπεριλαμβανομένων, ενδεικτικά, διευθύνσεων IP, μοναδικών αναγνωριστικών και δεδομένων περιήγησης που είναι αποθηκευμένα σε cookie, προκειμένου να επεξεργαστούμε προσωπικά δεδομένα. Έχετε την επιλογή να διαχειριστείτε τις προτιμήσεις σας επιλέγοντας την επιλογή «Διαχείριση προτιμήσεων» που βρίσκεται στο υποσέλιδο. Για να επιθεωρήσετε ή να εναντιωθείτε σε περιπτώσεις όπου οι συνεργάτες μας ισχυρίζονται ότι έχουν έννομο συμφέρον να χρησιμοποιήσουν τα δεδομένα σας, επισκεφτείτε τη σελίδα των προμηθευτών μας."
    },
    "it": {
        "purpose_header": "Noi e i nostri partner elaboriamo i dati per le finalità seguenti",
        "description": "Noi e i nostri __VENDOR_COUNT_LINK__ partner utilizziamo cookie e metodi analoghi per riconoscere i visitatori e ricordare le loro preferenze. Possiamo anche utilizzare queste tecnologie per verificare l’efficacia di campagne pubblicitarie, inviare pubblicità mirate, ed analizzare il traffico verso il sito. Alcune di queste tecnologie sono essenziali per assicurare il funzionamento del servizio o del sito e non possono essere disabilitate; altre sono opzionali ma migliorano l’esperienza degli utilizzatori in vari modi. \n\nInsieme ai nostri __VENDOR_COUNT_LINK__ partner, registriamo e/o leggiamo informazioni sul dispositivo dell’utente, compresi ma non limitate a: indirizzi IP, identificativi unici, e dati di navigazione conservati nei cookie, al fine di gestire i dati personali. Hai la possibilità di gestire le tue preferenze selezionando l'opzione \"Gestisci preferenze\" in fondo alla pagina. Per verificare o opporsi in quei casi in cui i nostri partner dichiarano un interesse legittimo nell'utilizzo dei tuoi dati, visita la nostra pagina sui fornitori."
    },
    "es": {
        "purpose_header": "Nosotros y nuestros socios tratamos datos para los siguientes fines",
        "description": "Nosotros y nuestros __VENDOR_COUNT_LINK__ socios utilizamos cookies y métodos similares para reconocer a los visitantes y recordar sus preferencias. También podemos utilizar estas tecnologías para medir la eficacia de las campañas publicitarias, orientar los anuncios de forma selectiva y analizar el tráfico del sitio web. Algunas de estas tecnologías son esenciales para garantizar el funcionamiento correcto del servicio o del sitio web y no se pueden desactivar, mientras que otras son opcionales pero sirven para mejorar la experiencia del usuario de diversas maneras. \n\nNosotros, en colaboración con nuestros __VENDOR_COUNT_LINK__ socios, almacenamos y/o accedemos a información en el dispositivo de un usuario, p. ej. direcciones IP, identificadores únicos y datos de navegación almacenados en cookies, con el fin de tratar datos personales. Tiene la opción de administrar sus preferencias seleccionando la opción «Administrar preferencias» que figura en el pie de página. Para revisar u oponerse a los casos en los que nuestros socios afirman tener un interés legítimo en utilizar sus datos, visite nuestra página de proveedores."
    },
    "ja": {
        "purpose_header": "当社およびパートナーは、以下の目的でデータを処理します。",
        "description": "当社および__VENDOR_COUNT_LINK__パートナーは、訪問者を認識し、その設定を記憶するためにクッキーや同様の方法を使用します。また、当社は、広告キャンペーンの効果を測定し、広告対象を絞り込み、ウェブサイトのトラフィックを分析するためにこれらの技術を使用します。これらの技術の一部は、サービスまたはウェブサイトが正しく機能するために必須であり、無効にすることができません。その他は選択可能ですが、様々な形でユーザー体験の向上に役立ちます。 \n\n当社は、__VENDOR_COUNT_LINK__パートナーと協力し、ユーザーのデバイスに情報を保存し、またそれらの情報にアクセスします。これらの情報には、個人データを処理するためクッキーに保存されるIPアドレス、一意識別子、閲覧データなどが含まれます。本ページ下部にある［Manage Preferences］オプションを選択すると、設定を変更できます。当社のパートナーが個人データ利用の正当な利益を主張する事例を検証するには、またはこれに異議を申し立てるには、当社のベンダーのページにアクセスしてください。"
    },
    "bs": {
        "purpose_header": "Mi i naši partneri obrađujemo podatke u sljedeće svrhe",
        "description": "Mi i naši __VENDOR_COUNT_LINK__ partneri koristimo kolačiće i slične metode kako bismo prepoznali posjetitelje i zapamtili njihove preferencije. Ove tehnologije možemo koristiti i za mjerenje učinkovitosti oglasnih kampanja, ciljanje oglasa i analizu saobraćaja na web lokaciji. Neke od ovih tehnologija bitne su za osiguravanje ispravnog rada usluge ili web stranice i ne mogu se onemogućiti, dok su druge opcionalne, ali služe za poboljšanje korisničkog iskustva na različite načine. \n\nMi, u suradnji s našim __VENDOR_COUNT_LINK__ partnerima, informacije pohranjujemo i/ili im pristupamo na korisnikovom uređaju, uključujući, ali ne ograničavajući se na IP adrese, jedinstvene identifikatore i podatke o pregledavanju pohranjene u kolačićima, u cilju obrade ličnih podataka. Imate mogućnost da upravljate svojim postavkama odabirom opcije 'Upravljanje postavkama' koja se nalazi u podnožju stranice. Posjetite stranicu naših dobavljača kako biste pregledali ili uložili prigovor na slučajeve u kojima naši partneri tvrde da imaju legitiman interes za korištenje vaših podataka."
    },
    "zh": {
        "purpose_header": "我们和我们的合作伙伴处理数据的目的如下",
        "description": "我们和 __VENDOR_COUNT_LINK__ 合作伙伴使用 cookie 和类似方法来识别访客并记住他们的偏好。我们还可能使用这些技术来衡量广告活动的有效性，定位广告并分析网站流量。其中一些技术对于确保服务或网站的正常运行至关重要，并且不能被禁用。其他技术则是可选的，但它们可以通过各种方式增强用户体验。 \n\n我们与 __VENDOR_COUNT_LINK__ 合作伙伴合作，在用户设备上存储和/或访问信息，包括但不限于 IP 地址、唯一标识符以及存储在 Cookie 中的浏览数据，以便处理个人数据。您可以通过选择页面页脚中的“管理首选项”选项来管理您的首选项。要查看或反对我们的合作伙伴主张使用您数据的合法权益情况，请访问我们的供应商页面。"
    },
    "bg": {
        "purpose_header": "Ние и нашите партньори обработваме данни за следните цели",
        "description": "Ние и нашите __VENDOR_COUNT_LINK__ партньори използваме „бисквитки“ и подобни методи, за да разпознаваме посетителите и да запомняме техните предпочитания. Възможно е също така да използваме тези технологии, за да измерваме ефективността на рекламните кампании, да насочваме реклами и да анализираме трафика на уебсайта. Някои от тези технологии са от съществено значение за осигуряване на правилното функциониране на услугата или уебсайта и не могат да бъдат деактивирани, докато други не са задължителни, но служат за подобряване на потребителското изживяване по различни начини. \n\nЗа да обработваме лични данни ние, в сътрудничество с нашите __VENDOR_COUNT_LINK__ партньори, съхраняваме и/или получаваме достъп до информация на устройството на потребителя, включително, но не само, IP адреси, уникални идентификатори и данни за сърфиране, съхранявани в бисквитки. Имате възможност да управлявате предпочитанията си, като изберете опцията „Управление на предпочитанията“, която се намира в долния колонтитул на страницата. За да прегледате или възразите срещу случаите, в които нашите партньори заявяват законен интерес да използват Вашите данни, моля, посетете страницата ни за доставчици."
    },
    "ar": {
        "purpose_header": "نقوم نحن وشركاؤنا بمعالجة البيانات للأغراض التالية",
        "description": "نحن وشركاؤنا __VENDOR_COUNT_LINK__ نستخدم ملفات تعريف الارتباط وطرقًا مماثلة للتعرف على الزوار وتذكر تفضيلاتهم. قد نستخدم أيضًا هذه التقنيات لقياس فعالية الحملات الإعلانية، واستهداف الإعلانات وتحليل حركة مرور الموقع الإلكتروني. بعض هذه التقنيات ضرورية ولا يمكن تعطيلها لضمان سير الخدمة أو الموقع الإلكتروني، وقد يكون بعضها الآخر اختياري وتعمل على تحسين تجربة المستخدم بطرق مختلفة. \n\nبالتعاون مع شركائنا __VENDOR_COUNT_LINK__، نخزّن و/أو نتيح الوصول إلى المعلومات الموجودة على جهاز المستخدم بما في ذلك على سبيل المثال لا الحصر: عناوين الآي بي، ومعرفات فريدة، وبيانات تصفح مخزنة في ملفات تعريف الارتباط، لمعالجة البيانات الشخصية. لديك خيار إدارة تفضيلاتك عن طريق تحديد خيار \"إدارة التفضيلات\" الموجود في أسفل الصفحة. للمراجعة أو الاعتراض على الحالات التي يؤكد فيها شركاؤنا على مصلحة مشروعة في استخدام بياناتك، يرجى زيارة صفحة البائعين لدينا."
    },
    "tr": {
        "purpose_header": "Biz ve ortaklarımız aşağıdaki amaçlarla verileri işleriz",
        "description": "Biz ve __VENDOR_COUNT_LINK__ ortaklarımız, ziyaretçileri tanımak ve tercihlerini hatırlamak için çerezleri ve benzer yöntemleri kullanırız. Bu teknolojileri ayrıca reklam kampanyalarının etkinliğini ölçmek, reklamları hedefe yönelik olarak göstermek ve web sitesi trafiğini analiz etmek için de kullanabiliriz. Bu teknolojilerden bazıları, hizmetin veya web sitesinin düzgün çalışmasını sağlamak için gereklidir ve devre dışı bırakılamaz; diğerleri ise isteğe bağlıdır ancak kullanıcı deneyimini çeşitli şekillerde geliştirmeye yarar. \n\n__VENDOR_COUNT_LINK__ ortaklarımızla iş birliği içinde, kişisel verileri işlemek için IP adresleri, benzersiz tanımlayıcılar ve çerezlerde depolanan tarama verileri dahil ancak bunlarla sınırlı olmamak üzere bir kullanıcının cihazındaki bilgileri depolar ve/veya bunlara erişiriz. Sayfanın alt kısmında bulunan “Tercihleri Yönet” seçeneğini belirleyerek tercihlerinizi yönetebilirsiniz. Ortaklarımızın verilerinizi kullanma konusunda meşru menfaat öne sürdüğü durumları incelemek veya bunlara itiraz etmek için lütfen satıcılar sayfamızı ziyaret edin."
    },
    "hu": {
        "purpose_header": "Partnereinkkel együttműködve a következő célokból kezelünk adatokat",
        "description": "__VENDOR_COUNT_LINK__ partnerünkkel együttműködve sütiket és hasonló módszereket használunk a látogatók felismerésére és beállításaik megjegyzésére. Ezeket a technológiákat a reklámkampányok hatékonyságának mérésére, a hirdetések célzására és a weboldal forgalmának elemzésére is használhatjuk. Ezen technológiák némelyike elengedhetetlen a szolgáltatás vagy a weboldal megfelelő működéséhez, és nem kapcsolható ki, míg mások opcionálisak, de különböző módon szolgálják a felhasználói élmény fokozását. \n\n__VENDOR_COUNT_LINK__ partnerünkkel együttműködve a személyes adatok feldolgozása érdekében információkat érünk el és/vagy tárolunk a felhasználó eszközén, beleértve, de nem kizárólagosan az IP-címeket, egyedi azonosítókat és a sütikben tárolt böngészési adatokat. Lehetősége van a beállítások kezelésére az oldal láblécében található „Beállítások kezelése” opció kiválasztásával. Az olyan esetek áttekintéséhez vagy kifogásolásához, amikor partnereink jogos érdeke fűződik az Ön adatainak felhasználásához, kérjük, látogasson el szállítói oldalunkra."
    },
    "pt-PT": {
        "purpose_header": "Nós e os nossos parceiros tratamos dados para as seguintes finalidades",
        "description": "Nós e os nossos parceiros __VENDOR_COUNT_LINK__utilizamos cookies e métodos semelhantes para reconhecer os visitantes e recordar as suas preferências. Também podemos utilizar estas tecnologias para avaliar a eficácia das campanhas publicitárias, direcionar anúncios e analisar o tráfego do website. Algumas destas tecnologias são essenciais para garantir o bom funcionamento do serviço ou do Web site e não podem ser desativadas, enquanto outras são opcionais, mas servem para melhorar a experiência do utilizador de várias formas. \n\nEm colaboração com os nossos parceiros __VENDOR_COUNT_LINK__, armazenamos e/ou acedemos a informações no dispositivo de um utilizador, incluindo, entre outros, endereços IP, identificadores únicos e dados de navegação armazenados em cookies, a fim de processar dados pessoais. Tem a opção de gerir as suas preferências selecionando a opção \"Gerir Preferências\" localizada no rodapé da página. Para rever ou opor-se a situações em que os nossos parceiros afirmam ter um interesse legítimo na utilização dos seus dados, visite a nossa página de fornecedores."
    },
    "sv": {
        "purpose_header": "Vi och våra partners behandlar uppgifter i följande syften",
        "description": "Vi och våra __VENDOR_COUNT_LINK__ partners använder kakor och liknande metoder för att känna igen besökare och komma ihåg deras inställningar. Vi kan också komma att använda dessa funktioner för att mäta effektiviteten av marknadsföringskampanjer, riktad marknadsföring och analysera webbplatsens trafik. Vissa av dessa funktioner är nödvändiga för att försäkra att webbplatsens tjänster fungerar som de ska och kan inte avaktiveras, medan andra är valfria men är till för att förbättra användarupplevelsen på olika sätt. \n\nVi, i samarbete med våra __VENDOR_COUNT_LINK__ partners, lagrar och/eller har tillgång till information på en användares enhet, inbegripet men inte begränsat till IP-adresser, unika identifieringskoder, och webbinformationen lagrad i kakor för att behandla personuppgifter. Du har möjligheten att hantera dina inställningar i menyn ”Hantera inställningar” längst ner på sidan. För att granska eller avböja instanser där våra partners hävdar berättigat intresse i användningen av din data, se våra leverantörers sidor."
    },
    "uk": {
        "purpose_header": "Ми та наші партнери обробляємо дані для вказаних нижче цілей",
        "description": "Ми та наші партнери __VENDOR_COUNT_LINK__ використовуємо файли cookie та подібні методи для розпізнавання відвідувачів і запам’ятовування їхніх вподобань. Ми також можемо використовувати ці технології для оцінки ефективності рекламних кампаній, таргетування реклами та аналізу відвідуваності вебсайтів. Деякі з цих технологій необхідні для забезпечення належного функціонування сервісу або вебсайту і не можуть бути відключені, тоді як інші є необов’язковими, проте різними способами слугують для покращення користувацького досвіду. \n\nУ співпраці з нашими партнерами __VENDOR_COUNT_LINK__ ми зберігаємо і/або отримуємо доступ до інформації на пристрої користувача, включаючи, зокрема, IP-адреси, унікальні ідентифікатори та дані перегляду, що зберігаються в файлах cookie, з метою обробки персональних даних. Ви можете керувати своїми налаштуваннями, вибравши опцію «Керувати налаштуваннями», розташовану в нижньому колонтитулі сторінки. Щоб переглянути або заперечити проти випадків, коли наші партнери заявляють про законний інтерес у використанні ваших даних, завітайте на сторінку наших постачальників."
    },
    "sr-Latn": {
        "purpose_header": "Mi i naši partneri obrađujemo podatke u sledeće svrhe",
        "description": "Mi i naši __VENDOR_COUNT_LINK__ partneri koristimo kolačiće i slične metode da bismo prepoznali posetioce i zapamtili njihove postavke. Ove tehnologije možemo da koristimo i za procenu efikasnosti reklamnih kampanja, ciljanje oglasa i analizu saobraćaja na veb-lokacijama. Neke od ovih tehnologija su od suštinskog značaja za obezbeđivanje pravilnog funkcionisanja usluge ili veb-lokacije i ne mogu se onemogućiti, dok su druge opcione, a služe za poboljšanje korisničkog iskustva na različite načine. \n\nU saradnji sa našim __VENDOR_COUNT_LINK__ partnerima, mi čuvamo informacije i/ili im pristupamo na korisnikovom uređaju, uključujući, između ostalog, IP adrese, jedinstvene identifikatore i podatke o pregledanju uskladištene u kolačićima u cilju obrade ličnih podataka. Imate mogućnost da upravljate svojim željenim postavkama tako što ćete izabrati opciju „Upravljanje podešavanjima“ koja se nalazi u podnožju stranice. Da biste pregledali ili uložili prigovor na slučajeve u kojima naši partneri tvrde da imaju legitiman interes za korišćenje vaših podataka, posetite našu stranicu dobavljača."
    },
    "sk": {
        "purpose_header": "My a naši partneri spracúvame údaje na tieto účely",
        "description": "My a naši __VENDOR_COUNT_LINK__ partneri používame súbory cookie a podobné metódy na rozpoznanie návštevníkov a zapamätanie ich predvolieb. Tieto technológie môžeme používať aj na meranie účinnosti reklamných kampaní, cielenie reklamy a analýzu návštevnosti webových stránok. Niektoré z týchto technológií sú nevyhnutné na zabezpečenie správneho fungovania služby alebo webových stránok a nemožno ich vypnúť, zatiaľ čo iné sú voliteľné, ale slúžia na zlepšenie používania rôznymi spôsobmi. \n\nV spolupráci s našimi __VENDOR_COUNT_LINK__ partnermi ukladáme a/alebo pristupujeme k informáciám v zariadení používateľa, okrem iného k adresám IP, jedinečným identifikátorom a údajom o prehliadaní uloženým v súboroch cookie, aby sme mohli spracúvať osobné údaje. Svoje predvoľby môžete spravovať výberom možnosti „Správa predvolieb“, ktorá sa nachádza v dolnej časti stránky. Ak chcete skontrolovať alebo namietať proti prípadom, keď naši partneri uplatňujú oprávnený záujem pri využívaní vašich údajov, navštívte stránku našich dodávateľov."
    },
    "ro": {
        "purpose_header": "Noi și partenerii noștri prelucrăm datele în următoarele scopuri",
        "description": "Noi și partenerii noștri __VENDOR_COUNT_LINK__ folosim module cookie și metode similare de recunoaștere a vizitatorilor și a reține preferințele lor. De asemenea, folosim aceste tehnologii pentru a cuantifica eficacitatea campaniilor de promovare, reclamele țintite și a analiza traficul de pe site-ul web. O parte dintre aceste tehnologii sunt esențiale pentru a garanta funcționarea corespunzătoare a serviciului sau a site-ului web și nu pot fi dezactivate, în timp ce altele sunt opționale, dar servesc la îmbunătățiraea experienței utilizatorului în diverse moduri. \n\nNoi, în colaborare cu partenerii noștri __VENDOR_COUNT_LINK__ , stocăm și/sau accesăm informațiile de pe un dispozitiv al utilizatorului, inclusiv, dar fără a se limita la adresele IP, identificatorii unici și explorarea datelor stocate în modulele cookie, pentru a prelucra datele dvs. cu caracter personal. Aveți opțiunea de a vă gestiona preferințele selectând opțiunea „Gestionare preferințe” situată în subsolul paginii. Pentru a revizui sau a vă opune instanțelor în care partenerii noștri declară un interes legitim vis-a-vis de utilizarea datelor dvs., vizitați pagina noastră de furnizori."
    },
    "pl": {
        "purpose_header": "My i nasi partnerzy przetwarzamy dane w niżej wymienionych celach",
        "description": "My i nasi partnerzy __VENDOR_COUNT_LINK__ wykorzystujemy pliki cookie i podobne metody do rozpoznawania odwiedzających i zapamiętywania ich preferencji. Możemy także wykorzystywać te technologie do mierzenia skuteczności kampanii reklamowych i reklam celowanych oraz analizowania ruchu w witrynie. Niektóre z tych technologii są konieczne do zapewnienia prawidłowego działania usługi lub witryny i nie mogą być dezaktywowane, a inne są opcjonalne ale służą do zapewnienia większego komfortu użytkownikowi na różne sposoby. \n\nWe współpracy z naszymi partnerami __VENDOR_COUNT_LINK__ przechowujemy informacje – w tym między innymi adresy IP, niepowtarzalne identyfikatory i dane z przeglądania stron zapisane w plikach cookie – i/ub uzyskujemy do nich dostęp na urządzeniu użytkownika, aby przetwarzać dane osobowe. Mają Państwo możliwość zarządzania swoimi preferencjami przez wybranie opcji „Zarządzaj preferencjami” dostępnej w stopce strony. W celu weryfikacji przypadków, w których nasi partnerzy domagają się uznania uzasadnionego interesu w wykorzystywaniu Państwa danych, lub sprzeciwienia się nim, można odwiedzić naszą stronę dla dostawców."
    },
    "lt": {
        "purpose_header": "Mes ir mūsų partneriai tvarkome duomenis šiais tikslais",
        "description": "Mes ir mūsų __VENDOR_COUNT_LINK__ partneriai naudojame slapukus ir panašius metodus, kad atpažintume lankytojus ir įsimintume jų nuostatas. Šias technologijas taip pat galime naudoti siekdami įvertinti reklamos kampanijų veiksmingumą, nukreipti reklamą ir analizuoti svetainės lankomumą. Kai kurios iš šių technologijų yra būtinos tinkamam paslaugos ar svetainės veikimui užtikrinti ir negali būti išjungtos, o kitos yra neprivalomos, tačiau padeda įvairiais būdais pagerinti naudotojo patirtį. \n\nMes, bendradarbiaudami su __VENDOR_COUNT_LINK__ partneriais, saugome ir (arba) prieiname prie naudotojo įrenginyje esančios informacijos, įskaitant, bet neapsiribojant, IP adresus, unikalius identifikatorius ir naršymo duomenis, saugomus slapukuose, kad galėtume tvarkyti asmens duomenis. Savo nuostatas galite tvarkyti pasirinkę puslapio poraštėje esančią parinktį „Tvarkyti nuostatas“. Jei norite peržiūrėti arba nesutikti su atvejais, kai mūsų partneriai nurodo teisėtą interesą naudoti jūsų duomenis, apsilankykite mūsų pardavėjų puslapyje."
    },
    "hr": {
        "purpose_header": "Mi i naši partneri obrađujemo podatke za sljedeće svrhe",
        "description": "Mi i naši __VENDOR_COUNT_LINK__ partneri koristimo kolačiće i slične metode kako bismo prepoznali posjetitelje i zapamtili njihove sklonosti. Ove tehnologije također možemo koristiti za procjenu učinkovitosti reklamnih kampanja, ciljanih oglasa i analizu prometa web-mjesta. Neke od ovih tehnologija ključne su za osiguravanje pravilnog funkcioniranja usluge ili web-mjesta i ne mogu biti onesposobljene, dok druge nisu obavezne, ali služe za poboljšanje korisničkog doživljaja na različite načine. \n\nMi, u suradnji s našim __VENDOR_COUNT_LINK__ partnerima, pohranjujemo i/ili pristupamo informacijama na korisnikovom uređaju, uključujući, ali ne ograničavajući se na IP adrese, jedinstvene identifikatore i pregledavanje podataka pohranjenih u kolačićima kako bismo obradili osobne podatke. Vi imate mogućnost upravljanja preferencijama odabirom opcije „Upravljanje preferencijama“ koja se nalazi u podnožju stranice. Da biste pregledali ili prigovorili na slučajeve kada naši partneri tvrde da imaju legitimni interes za korištenje vaših podataka, posjetite našu stranicu za pružatelje usluga."
    },
    "no": {
        "purpose_header": "Vi og partnerne våre håndterer data for følgende hensikter",
        "description": "Vi og __VENDOR_COUNT_LINK__ partnerne våre bruker informasjonskapsler og lignende metoder til å kjenne igjen besøkende og huske preferansene deres. Det kan hende at vi også bruker disse teknologiene til å måle effektiviteten til annonsekampanjer, målrette annonser og analysere nettstedtrafikk. Enkelte av disse teknologiene er nødvendige for å sikre at tjenesten eller nettstedet fungerer som de skal, og de kan derfor ikke slås av. Andre er valgfrie, men funksjonen deres er å forbedre brukeropplevelsen på ulike måter. \n\nVi, i samarbeid med __VENDOR_COUNT_LINK__ partnerne våre, lagrer og/eller får tilgang til informasjon på en brukers enhet. Dette inkluderer, men er ikke begrenset til, IP-adresser, unike identifikatorer og nettleserdata lagret i informasjonskapsler, sånn at disse kan behandle personlige data. Du har muligheten til å administrere preferansene dine ved å velge alternativet «Administrer preferanser» nederst på siden. Hvis du vil kontrollere eller rette en innsigelse i tilfeller hvor partnerne våre hevder en legitim interesse i å bruke dataene dine, finner du mer info om dette på leverandørsiden vår."
    },
    "da": {
        "purpose_header": "Vi og vores partnere behandler data til følgende formål",
        "description": "Vi og vores __VENDOR_COUNT_LINK__ partnere bruger cookies og lignende metoder til at genkende besøgende og huske deres præferencer. Vi kan også bruge disse teknologier til at måle reklamekampagners effektivitet, målrette reklamer og analysere webstedstrafik. Nogle af disse teknologier er afgørende for at sikre, at tjenesten eller webstedet fungerer korrekt, og at den/det ikke kan deaktiveres, mens andre er valgfrie, men har til formål at forbedre brugeroplevelsen på forskellige måder. \n\nVi, i samarbejde med vores __VENDOR_COUNT_LINK__ partnere, gemmer og/eller tager os adgang til oplysninger på en brugers enhed, herunder, men ikke begrænset til IP-adresser, entydige id'er og browsingdata, der gemmes i cookies, for at behandle persondata. Du har mulighed for at administrere dine præferencer ved at vælge muligheden \"Administrer præferencer\", der er at finde i sidens sidefod. For at gennemgå eller komme med indsigelser til tilfælde, hvor vores partnere hævder at have en legitim interesse i anvendelsen af dine data kan du besøge siden Vores forhandlere."
    },
    "ca": {
        "purpose_header": "Nosaltres i els nostres socis tractem les dades amb les finalitats següents",
        "description": "Nosaltres i els nostres socis de __VENDOR_COUNT_LINK__ utilitzem galetes i mètodes semblants per reconèixer els visitants i recordar-ne les preferències. També podem utilitzar aquestes tecnologies per avaluar l’efectivitat de les campanyes publicitàries, dirigir els anuncis i analitzar el trànsit del lloc web. Algunes d’aquestes tecnologies són essencials per garantir el bon funcionament del servei o lloc web i no es poden desactivar, mentre que d’altres són opcionals però serveixen per millorar l’experiència de l’usuari de diverses maneres. \n\nNosaltres, en col·laboració amb els nostres socis de __VENDOR_COUNT_LINK__, accedim a informació del dispositiu d’un usuari o l’emmagatzemem. Aquesta informació inclou, entre d’altres, adreces IP, identificadors únics i dades de navegació emmagatzemades a les galetes per processar les dades personals. Podeu gestionar les vostres preferències seleccionant l’opció “Gestiona les preferències” que hi ha al peu de la pàgina. Si voleu revisar o oposar-vos als casos en què els nostres socis afirmen tenir un interès legítim per utilitzar les vostres dades, visiteu la nostra pàgina de proveïdors."
    },
    "de": {
        "purpose_header": "Wir und unsere Partner verarbeiten Daten für folgende Zwecke",
        "description": "Wir und unsere __VENDOR_COUNT_LINK__ Partner verwenden Cookies und ähnliche Techniken, um Besucher wiederzuerkennen und ihre Einstellungen zu speichern. Wir können diese Technologien auch verwenden, um die Effektivität von Werbekampagnen zu messen, Werbung gezielt zu platzieren und den Traffic der Website zu analysieren. Einige dieser Technologien sind für die ordnungsgemäße Funktion des Dienstes oder der Website unerlässlich und können nicht deaktiviert werden. Andere sind optional, dienen aber dazu, die Benutzererfahrung auf verschiedene Weise zu verbessern. \n\nIn Zusammenarbeit mit unseren __VENDOR_COUNT_LINK__ Partnern speichern wir Informationen auf den Geräten der Benutzer und/oder greifen darauf zu. Zu diesen Informationen zählen unter anderem IP-Adressen, eindeutige Kennungen und in Cookies gespeicherte Browserdaten, um persönliche Daten zu verarbeiten. Sie können Ihre Einstellungen verwalten, indem Sie in der Fußzeile der Seite auf die Option „Einstellungen verwalten“ klicken. Gehen Sie bitte auf unsere Seite für Anbieter, um Fälle zu prüfen, in denen unsere Partner ein berechtigtes Interesse an der Nutzung Ihrer Daten geltend machen, oder um der Nutzung zu widersprechen."
    },
    "pt-BR": {
        "purpose_header": "Nós e nossos parceiros processamos dados para os seguintes fins",
        "description": "Nós e nossos __VENDOR_COUNT_LINK__ parceiros usamos cookies e métodos semelhantes para reconhecer visitantes e lembrar suas preferências. Também podemos usar essas tecnologias para avaliar a eficácia de campanhas publicitárias, direcionar anúncios e analisar o tráfego do site. Algumas destas tecnologias são essenciais para garantir o bom funcionamento do serviço ou do site e não podem ser desativadas, enquanto outras são opcionais, mas servem para melhorar a experiência do usuário de diversas formas. \n\nNós, em colaboração com nossos __VENDOR_COUNT_LINK__ parceiros, armazenamos e/ou acessamos informações no dispositivo de um usuário, incluindo, mas não se limitando a, endereços de IP, identificadores únicos e dados de navegação armazenados em cookies, com o objetivo de processar dados pessoais. Você tem a opção de gerenciar suas preferências selecionando a opção “Gerenciar Preferências” localizada no rodapé da página. Para analisar ou contestar casos em que nossos parceiros aleguem ter um interesse legítimo na utilização de seus dados, visite nossa página de fornecedores."
    },
    "fr": {
        "purpose_header": "Nos partenaires et nous-mêmes traitons les données aux fins suivantes",
        "description": "Nos __VENDOR_COUNT_LINK__ partenaires et nous-mêmes utilisons des cookies et des méthodes similaires pour reconnaître les visiteurs de ce site et nous souvenir de leurs préférences. Nous pouvons également utiliser ces technologies pour mesurer l’efficacité des campagnes publicitaires et des publicités ciblées, et pour analyser le trafic sur le site Web. Certaines de ces technologies sont indispensables au bon fonctionnement du service ou du site Web et ne peuvent pas être désactivées, tandis que d’autres sont facultatives, mais permettent d’améliorer l’expérience de l’utilisateur de multiples façons. \n\nEn collaboration avec nos __VENDOR_COUNT_LINK__ partenaires, nous stockons ou consultons des informations sur l’appareil d’un utilisateur, y compris, mais sans s’y limiter, les adresses IP, les identifiants uniques et les données de navigation stockées par les cookies, afin de traiter les renseignements personnels. Vous avez la possibilité de gérer vos préférences en sélectionnant l’option « Gérer les préférences » située en pied de page. Pour consulter ou refuser les instances pour lesquelles nos partenaires invoquent un intérêt légitime d’utilisation de vos données, veuillez vous reporter à la page de nos fournisseurs."
    },
    "gl": {
        "purpose_header": "Os nosos socios e nós procesamos datos para os seguintes fins",
        "description": "Os nosos __VENDOR_COUNT_LINK__ socios e nós utilizamos cookies e métodos similares para recoñecer os visitantes e recordar as súas preferencias. Tamén podemos usar estas tecnoloxías para avaliar a efectividade das campañas publicitarias, dirixir anuncios e analizar o tráfico da web. Algunhas son esenciais para garantir un bo funcionamento do servizo ou da páxina web e non poden desactivarse, mentres que outras son opcionais pero serven para mellorar a experiencia do usuario de diversas maneiras. \n\nNós, en  colaboración cos nosos __VENDOR_COUNT_LINK__ socios, almacenamos e/ou accedemos á información dos dispositivos do usuario como direccións IP, identificadores únicos e datos de navegación gardados nas cookies para procesar a información persoal. Vostede ten a posibilidade de xestionar as súas preferencias seleccionando a opción \"Xestionar preferencias\" que se atopa ao pé da páxina. Para revisar ou oporse aos casos nos que os nosos socios afirman ter un interese lexítimo para utilizar os seus datos, visite a páxina dos nosos subministradores."
    },
    "cs": {
        "purpose_header": "My a naši partneři zpracováváme data pro následující účely",
        "description": "My a naši __VENDOR_COUNT_LINK__ partneři užíváme soubory cookie a podobné metody k rozpoznání návštěvníků a zapamatování jejich preferencí. Tyto technologie můžeme také použít k měření účinnosti reklamních kampaní, cílení reklam a analýze návštěvnosti webových stránek. Některé z těchto technologií jsou nezbytné pro zajištění správného fungování služby nebo webových stránek a nelze je deaktivovat, zatímco jiné jsou volitelné, ale slouží různými způsoby k vylepšení uživatelského zážitku. \n\nVe spolupráci s našimi __VENDOR_COUNT_LINK__ partnery ukládáme a/nebo přistupujeme k informacím na zařízení uživatele, včetně IP adres, jedinečných identifikátorů a údajů o prohlížení uložených v souborech cookie, za účelem zpracování osobních údajů. Máte možnost spravovat své předvolby výběrem možnosti „Spravovat předvolby“ v zápatí stránky. Chcete-li zkontrolovat nebo odmítnou případy, kdy naši partneři uplatňují legitimní zájem na využití vašich údajů, navštivte naši stránku pro dodavatele."
    },
    "sr-Cyrl": {
        "purpose_header": "Ми и наши партнери обрађујемо податке у следеће сврхе",
        "description": "Ми и наши __VENDOR_COUNT_LINK__ партнери користимо колачиће и сличне методе за препознавање наших посетилаца и памћење њихових поставки. Такође можемо користити ове технологије за процену ефикасности рекламних кампања, циљање реклама и анализу саобраћаја на вебсајту. Неке од ових технологија су од суштинског значаја за обезбеђивање правилног функционисања услуге или вебсајта и не могу се онемогућити, док су друге изборне, али служе за побољшање корисничког искуства на различите начине. \n\nУ сарадњи са нашим __VENDOR_COUNT_LINK__ партнерима, ми чувамо информације и/или им приступамо на корисниковом уређају, што укључује, између осталог, IP адресе, јединствене идентификаторе и податке о прегледању сачуване у колачићима, како бисмо обрађивали личне податке. Имате опцију да управљате својим подешавањима тако што ћете изабрати опцију „Управљање подешавањима“ која се налази у подножју странице. Да бисте прегледали или уложили приговор на случајеве у којима наши партнери тврде да имају легитиман интерес за коришћење ваших података, посетите нашу страницу добављача."
    },
    "ru": {
        "purpose_header": "Мы с партнерами обрабатываем данные в следующих целях",
        "description": "Мы с партнерами __VENDOR_COUNT_LINK__ используем файлы cookie и аналогичные методы для распознавания посетителей и запоминания их предпочтений. Мы также можем использовать эти технологии для оценки эффективности рекламных кампаний, целевой рекламы и анализа посещаемости сайта. Некоторые из этих технологий необходимы для обеспечения надлежащего функционирования сервиса или сайта. Их нельзя отключить. Другие необязательны, но помогают улучшить пользовательский опыт различными способами. \n\nВ сотрудничестве со своими партнерами __VENDOR_COUNT_LINK__ мы храним информацию на устройстве пользователя и (или) обращаемся к ней для обработки персональных данных. К этой информации относятся, помимо прочего, IP-адреса, уникальные идентификаторы и данные о просмотре сайтов, хранящиеся в файлах cookie. Вы можете управлять своими настройками, выбрав «Управление настройками» в нижнем колонтитуле страницы. Узнать, в каких случаях наши партнеры заявляют о своем законном интересе в использовании ваших данных, или возразить против такого использования можно на странице, посвященной нашим поставщикам."
    },
    "sl": {
        "purpose_header": "Mi in naši partnerji obdelujemo podatke za naslednje namene",
        "description": "Mi in naši partnerji __VENDOR_COUNT_LINK__ uporabljamo piškotke in podobne metode, da prepoznamo obiskovalce in si zapomnimo njihove nastavitve. Te tehnologije lahko uporabljamo tudi za merjenje učinkovitosti oglaševalskih kampanj, ciljno usmerjanje oglasov in analizo prometa na spletnem mestu. Nekatere od teh tehnologij so nujne za zagotavljanje pravilnega delovanja storitve ali spletnega mesta in jih ni mogoče onemogočiti, medtem ko so druge neobvezne, vendar služijo za izboljšanje uporabniške izkušnje na različne načine. \n\nV sodelovanju z našimi partnerji __VENDOR_COUNT_LINK__ shranjujemo in/ali dostopamo do informacij na uporabnikovi napravi, kar med drugim vključuje naslove IP, edinstvene identifikatorje in podatke o brskanju, shranjene v piškotkih, z namenom obdelave osebnih podatkov. Svoje nastavitve lahko upravljate tako, da izberete možnost »Upravljanje nastavitev« v nogi strani. Če želite pregledati ali ugovarjati primerom, v katerih naši partnerji uveljavljajo zakoniti interes za uporabo vaših podatkov, obiščite našo stran o partnerjih."
    },
    "mt": {
        "purpose_header": "Aħna u l-imsieħba tagħna nipproċessaw id-dejta għall-iskopijiet li ġejjin",
        "description": "Aħna u l-imsieħba __VENDOR_COUNT_LINK__ tagħna nużaw il-cookies u metodi simili biex nagħrfu l-viżitaturi u niftakru l-preferenzi tagħhom. Nistgħu nużaw ukoll dawn it-teknoloġiji biex inkejlu l-effettività tal-kampanji ta’ reklamar, noħolqu reklami mmirati, u nanalizzaw in-numru ta’ żjarat fis-sit web. Xi wħud minn dawn it-teknoloġiji huma essenzjali biex jiġi żgurat il-funzjonament xieraq tas-servizz jew tas-sit web, b’hekk dawn ma jistgħux jiġu diżattivati, filwaqt li oħrajn mhumiex obbligatorji iżda jservu biex itejbu l-esperjenza tal-utent b’diversi modi. \n\nAħna, f’kollaborazzjoni mal-imsieħba __VENDOR_COUNT_LINK__ tagħna, naħżnu u/jew naċċessaw l-informazzjoni fuq l-apparat ta’ utent, li tinkludi iżda li mhijiex limitata għal indirizzi IP, identifikaturi uniċi, u dejta tal-ibbrawżjar maħżuna fil-cookies, sabiex nipproċessaw dejta personali. Għandek l-għażla li timmaniġġja l-preferenzi tiegħek billi tagħżel l-għażla ‘Immaniġġja l-Preferenzi’ li tinsab fil-qiegħ tal-paġna. Biex tirrevedi jew toġġezzjona għal każijiet fejn l-imsieħba tagħna jaffermaw interess leġittimu fl-użu tad-dejta tiegħek, jekk jogħġbok żur il-paġna tal-bejjiegħa tagħna."
    },
    "nl": {
        "purpose_header": "Wij en onze partners verwerken gegevens voor de volgende doeleinden",
        "description": "Wij en onze __VENDOR_COUNT_LINK__ partners gebruiken cookies en vergelijkbare methoden om bezoekers te herkennen en hun voorkeuren te onthouden. We kunnen deze technologieën ook gebruiken om de effectiviteit van advertentiecampagnes te meten, gericht te adverteren en websiteverkeer te analyseren. Sommige van deze technologieën zijn essentieel voor de juiste werking van de dienst of website en kunnen niet uitgeschakeld worden, terwijl andere optioneel zijn maar de gebruikerservaring op verschillende manieren verbeteren. \n\nWij, in samenwerking met onze __VENDOR_COUNT_LINK__ partners, bewaren en/of gebruiken informatie op het apparaat van een gebruiker, inclusief maar niet beperkt tot IP-adressen, unieke identificatiegegevens en in cookies bewaarde browsegegevens, om persoonlijke gegevens te verwerken. U kunt uw voorkeuren beheren door de optie 'Voorkeuren beheren' in de voettekst van de pagina te selecteren. Ga naar onze leverancierspagina als u wilt zien in welke gevallen onze partners zich beroepen op een gerechtvaardigd belang om uw gegevens te gebruiken of als u daar bezwaar tegen wilt maken."
    },
    "fi": {
        "purpose_header": "Me ja kumppanimme käsittelemme tietoja seuraaviin tarkoituksiin",
        "description": "Me ja __VENDOR_COUNT_LINK__ kumppanimme käytämme evästeitä ja vastaavia menetelmiä vierailijoiden tunnistamiseen ja heidän valintojensa muistamiseen. Voimme käyttää näitä tekniikoita myös mainoskampanjoiden tehokkuuden mittaamiseen, mainosten kohdistamiseen ja verkkosivuston liikenteen analysointiin. Jotkut näistä teknologioista ovat välttämättömiä palvelun tai verkkosivuston moitteettoman toiminnan varmistamiseksi, eikä niitä voi poistaa käytöstä. Toiset taas ovat valinnaisia, mutta ne parantavat käyttökokemusta eri tavoin. \n\nYhteistyössä __VENDOR_COUNT_LINK__ kumppaneidemme kanssa tallennamme käyttäjän laitteelle tietoja ja/tai käytämme niitä, mukaan lukien, mutta ei rajoittuen, IP-osoitteet, yksilölliset tunnisteet ja evästeisiin tallennetut selaustiedot henkilötietojen käsittelyä varten. Asetuksia voi hallita valitsemalla ”Hallinnoi asetuksia” -vaihtoehdon sivun alatunnisteessa. Toimittajasivullamme voi tarkastaa tai vastustaa tapauksia, joissa kumppanimme käyttävät oikeutettua etua tietojen hyödyntämiseen."
    },
    "en": {
        "purpose_header": "We and our partners process data for the following purposes",
        "description": "We and our __VENDOR_COUNT_LINK__ partners use cookies and similar methods to recognize visitors and remember their preferences. We may also use these technologies to gauge the effectiveness of advertising campaigns, target advertisements, and analyze website traffic. Some of these technologies are essential for ensuring the proper functioning of the service or website and cannot be disabled, while others are optional but serve to enhance the user experience in various ways. \n\nWe, in collaboration with our __VENDOR_COUNT_LINK__ partners, store and/or access information on a user's device, including but not limited to IP addresses, unique identifiers, and browsing data stored in cookies, in order to process personal data. You have the option to manage your preferences by selecting the 'Manage Preferences' option located in the page's footer. To review or object to instances where our partners assert a legitimate interest in utilizing your data, please visit our vendors page."
    }
}
# fmt: on


def upgrade():
    logger.info(
        f"Starting TCF Description Update and Addition of new Purpose Header {datetime.now()}"
    )

    bind = op.get_bind()

    # Getting details about the ExperienceTranslations along with their latest historical version
    # from ExperienceConfigHistory (every time a PrivacyExperienceConfig and/or an ExperienceTranslation is updated,
    # a new version is created in PrivacyExperienceConfigHistory).
    translation_and_latest_hist_join = bind.execute(
        text(
            """
        SELECT et.id as et_id, hist.id as hist_id, hist.version, et.language, et.description FROM experiencetranslation et
        JOIN privacyexperienceconfighistory hist ON et.id=hist.translation_id 
        WHERE hist.component = 'tcf_overlay'
        AND hist.version = (
            SELECT MAX(version) 
            FROM privacyexperienceconfighistory 
            WHERE translation_id=hist.translation_id
        );
        """
        )
    )

    for res in translation_and_latest_hist_join:
        if res["language"] in updated_translations:
            updated_translation = updated_translations[res["language"]]
            updated_translation["translation_id"] = res["et_id"]

            if res["version"] != 1:
                # Only update translation descriptions if there haven't been any edits
                updated_translation["description"] = res["description"]

            # For every ExperienceTranslation for which we have new data, add the purpose header,
            # in all cases, and update the description provided it's not a version 1
            bind.execute(
                text(
                    """
                    UPDATE experiencetranslation
                    SET description=:description, purpose_header=:purpose_header
                    WHERE id=:translation_id;
                    """
                ),
                updated_translation,
            )

            # Now to mimic what we do when a translation is updated, create a new ExperienceConfigHistory
            # record with the updated values and a bumped version.
            create_experience_config_history = text(
                """
                 INSERT INTO privacyexperienceconfighistory (
                        id,
                        component,
                        disabled,
                        is_default,
                        reject_button_label,
                        version,
                        accept_button_label,
                        acknowledge_button_label,
                        banner_enabled,
                        description, 
                        privacy_preferences_link_label,
                        privacy_policy_link_label,
                        privacy_policy_url,
                        save_button_label,
                        title, 
                        banner_description,
                        banner_title, 
                        language,
                        name,
                        origin,
                        dismissable,
                        auto_detect_language,
                        allow_language_selection,
                        translation_id,
                        modal_link_label,
                        show_layer1_notices,
                        layer1_button_options,
                        purpose_header
                 )
                 SELECT
                    :id,
                    component,
                    disabled,
                    is_default,
                    reject_button_label,
                    :version,
                    accept_button_label,
                    acknowledge_button_label,
                    banner_enabled,
                    :description, 
                    privacy_preferences_link_label,
                    privacy_policy_link_label,
                    privacy_policy_url,
                    save_button_label,
                    title, 
                    banner_description,
                    banner_title, 
                    language,
                    name,
                    origin,
                    dismissable,
                    auto_detect_language,
                    allow_language_selection,
                    translation_id,
                    modal_link_label,
                    show_layer1_notices,
                    layer1_button_options,
                    :purpose_header
                FROM privacyexperienceconfighistory
                WHERE id = :existing_historical_id;
                """
            )

            bind.execute(
                create_experience_config_history,
                {
                    "id": generate_record_id("pri"),
                    "version": res["version"] + 1,
                    "existing_historical_id": res["hist_id"],
                    "description": updated_translation["description"],
                    "purpose_header": updated_translation["purpose_header"],
                },
            )


def downgrade():
    pass
